"""sox.pyx - a thin wrapper around libsox"""

from types import SimpleNamespace
from cpython.buffer cimport PyObject_CheckBuffer, PyObject_GetBuffer, PyBuffer_Release, PyBUF_WRITABLE, PyBUF_C_CONTIGUOUS, Py_buffer
from cpython.ref cimport PyObject
from libc.stdint cimport uintptr_t

cimport cysox.sox


# Global storage for flow_effects callback
# We use a dictionary to support multiple concurrent chains (thread-safety handled by GIL)
_flow_callbacks = {}

# Storage for the last callback exception (since exceptions can't propagate through C code)
_last_callback_exception = None

# Track initialization state to make init()/quit() idempotent
# (libsox crashes on repeated init/quit cycles, so we only do it once)
_sox_initialized = False


ERROR_CODES = {
    'SOX_SUCCESS': SOX_SUCCESS,
    'SOX_EOF': SOX_EOF,
    'SOX_EHDR': SOX_EHDR,
    'SOX_EFMT': SOX_EFMT,
    'SOX_ENOMEM': SOX_ENOMEM,
    'SOX_EPERM': SOX_EPERM,
    'SOX_ENOTSUP': SOX_ENOTSUP,
    'SOX_EINVAL': SOX_EINVAL,
}

# Expose constants at module level - Python accessible
SUCCESS = SOX_SUCCESS
EOF = SOX_EOF
EHDR = SOX_EHDR
EFMT = SOX_EFMT
ENOMEM = SOX_ENOMEM
EPERM = SOX_EPERM
ENOTSUP = SOX_ENOTSUP
EINVAL = SOX_EINVAL


ENCODINGS = [
    ("UNKNOWN",     "encoding has not yet been determined"),
    ("SIGN2",       "signed linear 2's comp: Mac"),
    ("UNSIGNED",    "unsigned linear: Sound Blaster"),
    ("FLOAT",       "floating point (binary format)"),
    ("FLOAT_TEXT",  "floating point (text format)"),
    ("FLAC",        "FLAC compression"),
    ("HCOM",        "Mac FSSD files with Huffman compression"),
    ("WAVPACK",     "WavPack with integer samples"),
    ("WAVPACKF",    "WavPack with float samples"),
    ("ULAW",        "u-law signed logs: US telephony, SPARC"),
    ("ALAW",        "A-law signed logs: non-US telephony, Psion"),
    ("G721",        "G.721 4-bit ADPCM"),
    ("G723",        "G.723 3 or 5 bit ADPCM"),
    ("CL_ADPCM  ",  "Creative Labs 8 --> 2,3,4 bit Compressed PCM"),
    ("CL_ADPCM16",  "Creative Labs 16 --> 4 bit Compressed PCM"),
    ("MS_ADPCM  ",  "Microsoft Compressed PCM"),
    ("IMA_ADPCM ",  "IMA Compressed PCM"),
    ("OKI_ADPCM ",  "Dialogic/OKI Compressed PCM"),
    ("DPCM",        "Differential PCM: Fasttracker 2 (xi)"),
    ("DWVW ",       "Delta Width Variable Word"),
    ("DWVWN",       "Delta Width Variable Word N-bit"),
    ("GSM",         "GSM 6.10 33byte frame lossy compression"),
    ("MP3",         "MP3 compression"),
    ("VORBIS",      "Vorbis compression"),
    ("AMR_WB",      "AMR-WB compression"),
    ("AMR_NB",      "AMR-NB compression"),
    ("CVSD",        "Continuously Variable Slope Delta modulation"),
    ("LPC10",       "Linear Predictive Coding"),
    ("OPUS",        "Opus compression"),
]

constant = SimpleNamespace(**{
    'INT8_MAX': SOX_INT8_MAX,
    'INT16_MAX': SOX_INT16_MAX,
    'INT24_MAX': SOX_INT24_MAX,
    'SAMPLE_PRECISION': SOX_SAMPLE_PRECISION,
    'SAMPLE_MAX': SOX_SAMPLE_MAX,
    'SAMPLE_MIN': SOX_SAMPLE_MIN,
    'SAMPLE_NEG': SOX_SAMPLE_NEG,

    'SIZE_MAX': SOX_SIZE_MAX,
    'UNSPEC': SOX_UNSPEC,
    'UNKNOWN_LEN': SOX_UNKNOWN_LEN,
    'IGNORE_LENGTH': SOX_IGNORE_LENGTH,
    'DEFAULT_RATE': SOX_DEFAULT_RATE,
    'DEFAULT_PRECISION': SOX_DEFAULT_PRECISION,
    'DEFAULT_ENCODING': SOX_DEFAULT_ENCODING,
    'LOOP_NONE': SOX_LOOP_NONE,
    'LOOP_8': SOX_LOOP_8,
    'LOOP_SUSTAIN_DECAY': SOX_LOOP_SUSTAIN_DECAY,
    'MAX_NLOOPS': SOX_MAX_NLOOPS,

    'FILE_NOSTDIO': SOX_FILE_NOSTDIO,
    'FILE_DEVICE': SOX_FILE_DEVICE,

    'FILE_PHONY': SOX_FILE_PHONY,
    'FILE_REWIND': SOX_FILE_REWIND,
    'FILE_BIT_REV': SOX_FILE_BIT_REV,
    'FILE_NIB_REV': SOX_FILE_NIB_REV,
    'FILE_ENDIAN': SOX_FILE_ENDIAN,
    'FILE_ENDBIG': SOX_FILE_ENDBIG,
    'FILE_MONO': SOX_FILE_MONO,
    'FILE_STEREO': SOX_FILE_STEREO,
    'FILE_QUAD': SOX_FILE_QUAD,

    'FILE_CHANS': SOX_FILE_CHANS,
    'FILE_LIT_END': SOX_FILE_LIT_END,
    'FILE_BIG_END': SOX_FILE_BIG_END,

    'EFF_CHAN': SOX_EFF_CHAN,
    'EFF_RATE': SOX_EFF_RATE,
    'EFF_PREC': SOX_EFF_PREC,
    'EFF_LENGTH': SOX_EFF_LENGTH,
    'EFF_MCHAN': SOX_EFF_MCHAN,
    'EFF_NULL': SOX_EFF_NULL,
    'EFF_DEPRECATED': SOX_EFF_DEPRECATED,
    'EFF_GAIN': SOX_EFF_GAIN,
    'EFF_MODIFY': SOX_EFF_MODIFY,
    'EFF_ALPHA': SOX_EFF_ALPHA,
    'EFF_INTERNAL': SOX_EFF_INTERNAL,
    'SEEK_SET': SOX_SEEK_SET,
})


# Custom exception classes for better error handling
class SoxError(Exception):
    """Base exception for all SoX-related errors."""
    pass


class SoxInitError(SoxError):
    """Exception raised when SoX initialization fails."""
    pass


class SoxFormatError(SoxError):
    """Exception raised when format operations fail."""
    pass


class SoxEffectError(SoxError):
    """Exception raised when effect operations fail."""
    pass


class SoxIOError(SoxError):
    """Exception raised when I/O operations fail."""
    pass


class SoxMemoryError(SoxError):
    """Exception raised when memory allocation fails."""
    pass



cdef class SignalInfo:
    """Signal parameters describing audio properties.

    This class encapsulates the signal characteristics of an audio stream,
    including sample rate, number of channels, precision, and length.

    All members can be set to SOX_UNSPEC (= 0) if unknown.

    Attributes:
        rate: Sample rate in samples per second (Hz), 0 if unknown
        channels: Number of audio channels, 0 if unknown
        precision: Bits per sample, 0 if unknown
        length: Total samples * channels in file, 0 if unknown, -1 if unspecified
        mult: Effects headroom multiplier, may be None

    Example:
        >>> signal = sox.SignalInfo(rate=44100, channels=2, precision=16)
        >>> f = sox.Format('output.wav', signal=signal, mode='w')
    """
    cdef sox_signalinfo_t* ptr
    cdef bint owner

    def __cinit__(self):
        self.ptr = NULL
        self.owner = False

    def __dealloc__(self):
        if self.ptr is not NULL and self.owner is True:
            if self.ptr.mult is not NULL:
                free(self.ptr.mult)
            free(self.ptr)
            self.ptr = NULL

    def __init__(self, rate: float = 0.0, channels: int = 0,
                precision: int = 0, length: int = 0, mult: float = 0.0):
        self.ptr = <sox_signalinfo_t*>malloc(sizeof(sox_signalinfo_t))
        if self.ptr is NULL:
            raise SoxMemoryError("Failed to allocate SignalInfo")
        self.rate = rate
        self.channels = channels
        self.precision = precision
        self.length = length
        self.mult = mult
        self.owner = True

    @staticmethod
    cdef SignalInfo from_ptr(sox_signalinfo_t* ptr, bint owner=False):
        cdef SignalInfo wrapper = SignalInfo.__new__(SignalInfo)
        wrapper.ptr = ptr
        wrapper.owner = owner
        return wrapper

    @property
    def rate(self) -> sox_rate_t:
        """samples per second, 0 if unknown"""
        if self.ptr is NULL:
            raise RuntimeError("SignalInfo pointer is NULL")
        return self.ptr.rate

    @rate.setter
    def rate(self, sox_rate_t value):
        if self.ptr is NULL:
            raise RuntimeError("SignalInfo pointer is NULL")
        self.ptr.rate = value

    @property
    def channels(self) -> int:
        """number of sound channels, 0 if unknown"""
        if self.ptr is NULL:
            raise RuntimeError("SignalInfo pointer is NULL")
        return self.ptr.channels

    @channels.setter
    def channels(self, int value):
        if self.ptr is NULL:
            raise RuntimeError("SignalInfo pointer is NULL")
        self.ptr.channels = value

    @property
    def precision(self) -> int:
        """bits per sample, 0 if unknown"""
        if self.ptr is NULL:
            raise RuntimeError("SignalInfo pointer is NULL")
        return self.ptr.precision

    @precision.setter
    def precision(self, int value):
        if self.ptr is NULL:
            raise RuntimeError("SignalInfo pointer is NULL")
        self.ptr.precision = value

    @property
    def length(self) -> sox_uint64_t:
        """samples * chans in file, 0 if unknown, -1 if unspecified"""
        if self.ptr is NULL:
            raise RuntimeError("SignalInfo pointer is NULL")
        return self.ptr.length

    @length.setter
    def length(self, sox_uint64_t value):
        if self.ptr is NULL:
            raise RuntimeError("SignalInfo pointer is NULL")
        self.ptr.length = value

    @property
    def mult(self) -> float:
        """Effects headroom multiplier may be null"""
        if self.ptr is NULL:
            raise RuntimeError("SignalInfo pointer is NULL")
        if not self.ptr.mult:
            return 0.0
        return self.ptr.mult[0]

    @mult.setter
    def mult(self, float value):
        if self.ptr is NULL:
            raise RuntimeError("SignalInfo pointer is NULL")
        if value == 0.0:
            # Free existing allocation before setting to NULL
            if self.ptr.mult is not NULL:
                free(self.ptr.mult)
            self.ptr.mult = NULL
        elif self.ptr.mult is NULL:
            self.ptr.mult = <double*>malloc(sizeof(double))
            if self.ptr.mult is NULL:
                raise SoxMemoryError("Failed to allocate memory for mult")
            self.ptr.mult[0] = value
        else:
            self.ptr.mult[0] = value


class EncodingsInfo:
    """Basic information about an encoding."""

    def __init__(self, sox_encodings_flags_t flags, str name, str desc):
        self.flags = flags
        self.name = name
        self.desc = desc

    def __repr__(self):
        return f"<EncodingsInfo '{self.name}' '{self.desc}' '{self.type}'>"

    @property
    def type(self):
        """Encoding type as human-readable string (lossless, lossy1, or lossy2)."""
        return {
            0: 'lossless',
            1: 'lossy1',
            2: 'lossy2',
        }[self.flags]


cdef class EncodingInfo:
    cdef sox_encodinginfo_t* ptr
    cdef bint owner

    def __cinit__(self):
        self.ptr = NULL
        self.owner = False

    def __dealloc__(self):
        if self.ptr is not NULL and self.owner is True:
            free(self.ptr)
            self.ptr = NULL

    def __init__(self, encoding: int = 0, bits_per_sample: int = 0,
                compression: float = 0.0, reverse_bytes: int = 0,
                reverse_nibbles: int = 0, reverse_bits: int = 0,
                opposite_endian: bool = False):
        self.ptr = <sox_encodinginfo_t*>malloc(sizeof(sox_encodinginfo_t))
        if self.ptr is NULL:
            raise SoxMemoryError("Failed to allocate EncodingInfo")
        self.encoding = encoding
        self.bits_per_sample = bits_per_sample
        self.compression = compression
        self.reverse_bytes = reverse_bytes
        self.reverse_nibbles = reverse_nibbles
        self.reverse_bits = reverse_bits
        self.opposite_endian = opposite_endian
        self.owner = True

    # def __init__(self):
    #     """Fills in an encodinginfo with default values."""
    #     self.ptr = <sox_encodinginfo_t*>malloc(sizeof(sox_encodinginfo_t))
    #     self.owner = True
    #     sox_init_encodinginfo(self.ptr)

    @staticmethod
    cdef EncodingInfo from_ptr(const sox_encodinginfo_t* ptr, bint owner=False):
        cdef EncodingInfo wrapper = EncodingInfo.__new__(EncodingInfo)
        wrapper.ptr = <sox_encodinginfo_t*>ptr
        wrapper.owner = owner
        return wrapper

    @property
    def encoding(self) -> sox_encoding_t:
        """format of sample numbers"""
        if self.ptr is NULL:
            raise RuntimeError("EncodingInfo pointer is NULL")
        return self.ptr.encoding

    @encoding.setter
    def encoding(self, sox_encoding_t value):
        if self.ptr is NULL:
            raise RuntimeError("EncodingInfo pointer is NULL")
        self.ptr.encoding = value

    @property
    def bits_per_sample(self) -> int:
        """0 if unknown or variable uncompressed value if lossless compressed value if lossy"""
        if self.ptr is NULL:
            raise RuntimeError("EncodingInfo pointer is NULL")
        return self.ptr.bits_per_sample

    @bits_per_sample.setter
    def bits_per_sample(self, unsigned value):
        if self.ptr is NULL:
            raise RuntimeError("EncodingInfo pointer is NULL")
        self.ptr.bits_per_sample = value

    @property
    def compression(self) -> float:
        """compression factor (where applicable)"""
        if self.ptr is NULL:
            raise RuntimeError("EncodingInfo pointer is NULL")
        return self.ptr.compression

    @compression.setter
    def compression(self, double value):
        if self.ptr is NULL:
            raise RuntimeError("EncodingInfo pointer is NULL")
        self.ptr.compression = value

    @property
    def reverse_bytes(self) -> sox_option_t:
        """Should bytes be reversed?"""
        if self.ptr is NULL:
            raise RuntimeError("EncodingInfo pointer is NULL")
        return self.ptr.reverse_bytes

    @reverse_bytes.setter
    def reverse_bytes(self, sox_option_t value):
        if self.ptr is NULL:
            raise RuntimeError("EncodingInfo pointer is NULL")
        self.ptr.reverse_bytes = value

    @property
    def reverse_nibbles(self) -> sox_option_t:
        """Should nibbles be reversed?"""
        if self.ptr is NULL:
            raise RuntimeError("EncodingInfo pointer is NULL")
        return self.ptr.reverse_nibbles

    @reverse_nibbles.setter
    def reverse_nibbles(self, sox_option_t value):
        if self.ptr is NULL:
            raise RuntimeError("EncodingInfo pointer is NULL")
        self.ptr.reverse_nibbles = value

    @property
    def reverse_bits(self) -> sox_option_t:
        """Should bits be reversed?"""
        if self.ptr is NULL:
            raise RuntimeError("EncodingInfo pointer is NULL")
        return self.ptr.reverse_bits

    @reverse_bits.setter
    def reverse_bits(self, sox_option_t value):
        if self.ptr is NULL:
            raise RuntimeError("EncodingInfo pointer is NULL")
        self.ptr.reverse_bits = value

    @property
    def opposite_endian(self) -> bool:
        """If set to true, the format should reverse its default endianness."""
        if self.ptr is NULL:
            raise RuntimeError("EncodingInfo pointer is NULL")
        return self.ptr.opposite_endian

    @opposite_endian.setter
    def opposite_endian(self, sox_bool value):
        if self.ptr is NULL:
            raise RuntimeError("EncodingInfo pointer is NULL")
        self.ptr.opposite_endian = value


cdef class LoopInfo:
    cdef sox_loopinfo_t* ptr
    cdef bint owner

    def __cinit__(self):
        self.ptr = NULL
        self.owner = False

    def __dealloc__(self):
        if self.ptr is not NULL and self.owner is True:
            free(self.ptr)
            self.ptr = NULL

    def __init__(self, start: int = 0, length: int = 0, count: int = 0,
                 type: int = 0):
        self.ptr = <sox_loopinfo_t*>malloc(sizeof(sox_loopinfo_t))
        self.start = start
        self.length = length
        self.count = count
        self.type = type
        self.owner = True

    @staticmethod
    cdef LoopInfo from_ptr(sox_loopinfo_t* ptr, bint owner=False):
        cdef LoopInfo wrapper = LoopInfo.__new__(LoopInfo)
        wrapper.ptr = ptr
        wrapper.owner = owner
        return wrapper

    @property
    def start(self) -> sox_uint64_t:
        """first sample"""
        if self.ptr is NULL:
            raise RuntimeError("LoopInfo pointer is NULL")
        return self.ptr.start

    @start.setter
    def start(self, sox_uint64_t value):
        if self.ptr is NULL:
            raise RuntimeError("LoopInfo pointer is NULL")
        self.ptr.start = value

    @property
    def length(self) -> sox_uint64_t:
        """length"""
        if self.ptr is NULL:
            raise RuntimeError("LoopInfo pointer is NULL")
        return self.ptr.length

    @length.setter
    def length(self, sox_uint64_t value):
        if self.ptr is NULL:
            raise RuntimeError("LoopInfo pointer is NULL")
        self.ptr.length = value

    @property
    def count(self) -> int:
        """number of repeats, 0=forever"""
        if self.ptr is NULL:
            raise RuntimeError("LoopInfo pointer is NULL")
        return self.ptr.count

    @count.setter
    def count(self, unsigned value):
        if self.ptr is NULL:
            raise RuntimeError("LoopInfo pointer is NULL")
        self.ptr.count = value

    @property
    def type(self) -> int:
        """0=no, 1=forward, 2=forward/back (see sox_loop_* for valid values)"""
        if self.ptr is NULL:
            raise RuntimeError("LoopInfo pointer is NULL")
        return self.ptr.type

    @type.setter
    def type(self, unsigned char value):
        if self.ptr is NULL:
            raise RuntimeError("LoopInfo pointer is NULL")
        self.ptr.type = value


cdef class InstrInfo:
    cdef sox_instrinfo_t* ptr
    cdef bint owner

    def __cinit__(self):
        self.ptr = NULL
        self.owner = False

    def __dealloc__(self):
        if self.ptr is not NULL and self.owner is True:
            free(self.ptr)
            self.ptr = NULL

    def __init__(self, note: int = 0, low: int = 0, high: int = 0,
                 loopmode: int = 0, nloops: int = 0):
        self.ptr = <sox_instrinfo_t*>malloc(sizeof(sox_instrinfo_t))
        self.note = note
        self.low = low
        self.high = high
        self.loopmode = loopmode
        self.nloops = nloops
        self.owner = True

    @staticmethod
    cdef InstrInfo from_ptr(sox_instrinfo_t* ptr, bint owner=False):
        cdef InstrInfo wrapper = InstrInfo.__new__(InstrInfo)
        wrapper.ptr = ptr
        wrapper.owner = owner
        return wrapper

    @property
    def note(self) -> int:
        """for unity pitch playback"""
        return self.ptr.MIDInote

    @note.setter
    def note(self, signed char value):
        self.ptr.MIDInote = value

    @property
    def low(self) -> int:
        """MIDI pitch-bend low range"""
        return self.ptr.MIDIlow

    @low.setter
    def low(self, signed char value):
        self.ptr.MIDIlow = value

    @property
    def high(self) -> int:
        """MIDI pitch-bend high range"""
        return self.ptr.MIDIhi

    @high.setter
    def high(self, signed char value):
        self.ptr.MIDIhi = value

    @property
    def loopmode(self) -> int:
        """0=no, 1=forward, 2=forward/back (see sox_loop_* values)"""
        return self.ptr.loopmode

    @loopmode.setter
    def loopmode(self, unsigned char value):
        self.ptr.loopmode = value

    @property
    def nloops(self) -> int:
        """number of active loops (max SOX_MAX_NLOOPS)"""
        return self.ptr.nloops

    @nloops.setter
    def nloops(self, unsigned value):
        self.ptr.nloops = value


cdef class FileInfo:
    cdef sox_fileinfo_t* ptr
    cdef bint owner

    def __cinit__(self):
        self.ptr = NULL
        self.owner = False

    def __dealloc__(self):
        if self.ptr is not NULL and self.owner is True:
            free(self.ptr)
            self.ptr = NULL

    def __init__(self, buf: bytes = None, size: int = 0, count: int = 0,
                 pos: int = 0):
        self.ptr = <sox_fileinfo_t*>malloc(sizeof(sox_fileinfo_t))
        if self.ptr is NULL:
            raise SoxMemoryError("Failed to allocate FileInfo")

        # Initialize struct members directly (buf property is read-only)
        # Note: buf parameter is ignored since setting it would be unsafe
        self.ptr.buf = NULL
        self.ptr.size = size
        self.ptr.count = count
        self.ptr.pos = pos
        self.owner = True

    @staticmethod
    cdef FileInfo from_ptr(sox_fileinfo_t* ptr, bint owner=False):
        cdef FileInfo wrapper = FileInfo.__new__(FileInfo)
        wrapper.ptr = ptr
        wrapper.owner = owner
        return wrapper

    @property
    def buf(self) -> bytes:
        """Pointer to data buffer (read-only).

        Note: This property is read-only because the buffer is managed internally
        by the sox_fileinfo_t structure. Setting it would create a dangling pointer
        to a temporary Python bytes object.
        """
        if self.ptr.buf == NULL:
            return None
        return self.ptr.buf[:self.ptr.size]

    @property
    def size(self) -> int:
        """Size of buffer in bytes"""
        return self.ptr.size

    @size.setter
    def size(self, size_t value):
        self.ptr.size = value

    @property
    def count(self) -> int:
        """Count read into buffer"""
        return self.ptr.count

    @count.setter
    def count(self, size_t value):
        self.ptr.count = value

    @property
    def pos(self) -> int:
        """Position in buffer"""
        return self.ptr.pos

    @pos.setter
    def pos(self, size_t value):
        self.ptr.pos = value


cdef class OutOfBand:
    cdef sox_oob_t* ptr
    cdef bint owner

    def __cinit__(self):
        self.ptr = NULL
        self.owner = False

    def __dealloc__(self):
        if self.ptr is not NULL and self.owner is True:
            # Clean up comments before freeing the struct
            if self.ptr.comments:
                sox_delete_comments(&self.ptr.comments)
            free(self.ptr)
            self.ptr = NULL

    def __init__(self):
        self.ptr = <sox_oob_t*>malloc(sizeof(sox_oob_t))
        self.owner = True

    @staticmethod
    cdef OutOfBand from_ptr(sox_oob_t* ptr, bint owner=False):
        cdef OutOfBand wrapper = OutOfBand.__new__(OutOfBand)
        wrapper.ptr = ptr
        wrapper.owner = owner
        return wrapper

    def num_comments(self) -> int:
        """Returns the number of items in the metadata block."""
        return <int>sox_num_comments(self.ptr.comments)

    def append_comment(self, item: str):
        """Adds an 'id=value' item to the metadata block."""
        sox_append_comment(&self.ptr.comments, item.encode())

    def append_comments(self, items: str):
        """Adds a newline-delimited list of "id=value" items to the metadata block."""
        sox_append_comments(&self.ptr.comments, items.encode())

    cdef sox_comments_t copy_comments(self):
        """Duplicates the metadata block."""
        return sox_copy_comments(self.ptr.comments)

    def find_comment(self, id: str) -> str:
        """If "id=value" is found, return value, else return null."""
        cdef char * value = <char*>sox_find_comment(self.ptr.comments, id.encode())
        if value is not NULL:
            return value.decode()

    @property
    def comments(self) -> list[str]:
        """Comment strings in id=value format."""
        cdef size_t n = sox_num_comments(self.ptr.comments)
        _comments = []
        for i in range(n):
            _comments.append(self.ptr.comments[i].decode())
        return _comments

    # FIXME: use list[str]
    # @comments.setter
    # def comments(self, sox_comments_t value):
    #     self.ptr.comments = value

    # cdef char** create_char_array(list python_strings):
    #     cdef int num_strings = len(python_strings)
    #     cdef char** c_array = <char**> malloc(num_strings * sizeof(char*))
    #     if not c_array:
    #         # Handle allocation error
    #         pass

    #     for i in range(num_strings):
    #         # Convert Python string to C string and store pointer
    #         c_array[i] = python_strings[i].encode('utf-8')  # Or other encoding
    #     return c_array

    # cdef void free_char_array(char** c_array, int num_strings):
    #     for i in range(num_strings):
    #         free(c_array[i]) # Free individual strings if they were allocated
    #     free(c_array)

    @property
    def instr(self) -> InstrInfo:
        """Instrument specification"""
        return InstrInfo.from_ptr(&self.ptr.instr, False)

    @property
    def loops(self) -> list:
        """Looping specification"""
        result = []
        for i in range(8):
            result.append(LoopInfo.from_ptr(&self.ptr.loops[i], False))
        return result


cdef class VersionInfo:
    cdef sox_version_info_t* ptr
    cdef bint owner

    def __cinit__(self):
        self.ptr = NULL
        self.owner = False

    def __dealloc__(self):
        if self.ptr is not NULL and self.owner is True:
            free(self.ptr)
            self.ptr = NULL

    @staticmethod
    cdef VersionInfo from_ptr(sox_version_info_t* ptr, bint owner=False):
        cdef VersionInfo wrapper = VersionInfo.__new__(VersionInfo)
        wrapper.ptr = ptr
        wrapper.owner = owner
        return wrapper

    @property
    def size(self) -> int:
        """structure size = sizeof(sox_version_info_t)"""
        return self.ptr.size

    @property
    def flags(self) -> sox_version_flags_t:
        """feature flags = popen | magic | threads | memopen"""
        return self.ptr.flags

    @property
    def version_code(self) -> sox_uint32_t:
        """version number = 0x140400"""
        return self.ptr.version_code

    @property
    def version(self) -> str:
        """version string = sox_version(), for example, 14.4.0"""
        if self.ptr.version == NULL:
            return None
        return self.ptr.version.decode()

    @property
    def version_extra(self) -> str:
        """version extra info or null = PACKAGE_EXTRA, for example, beta"""
        if self.ptr.version_extra == NULL:
            return None
        return self.ptr.version_extra.decode()

    @property
    def time(self) -> str:
        """build time = __DATE__ __TIME__, for example, Jan  7 2010 03:31:50"""
        if self.ptr.time == NULL:
            return None
        return self.ptr.time.decode()

    @property
    def distro(self) -> str:
        """distro or null = DISTRO, for example, Debian"""
        if self.ptr.distro == NULL:
            return None
        return self.ptr.distro.decode()

    @property
    def compiler(self) -> str:
        """compiler info or null, for example, msvc 160040219"""
        if self.ptr.compiler == NULL:
            return None
        return self.ptr.compiler.decode()

    @property
    def arch(self) -> str:
        """arch, for example, 1248 48 44 L OMP"""
        if self.ptr.arch == NULL:
            return None
        return self.ptr.arch.decode()


cdef class Globals:
    cdef sox_globals_t* ptr
    cdef bint owner

    def __cinit__(self):
        self.ptr = NULL
        self.owner = False

    def __dealloc__(self):
        if self.ptr is not NULL and self.owner is True:
            free(self.ptr)
            self.ptr = NULL

    def __init__(self):
        self.ptr = sox_get_globals()
        self.owner = False # not owned
        if self.ptr is NULL:
            raise MemoryError

    @staticmethod
    cdef Globals from_ptr(sox_globals_t* ptr, bint owner=False):
        cdef Globals wrapper = Globals.__new__(Globals)
        wrapper.ptr = ptr
        wrapper.owner = owner
        return wrapper

    def as_dict(self):
        """Returns a dict of libSoX's global settings."""
        return {
            'verbosity': self.verbosity,
            'repeatable': self.repeatable,
            'bufsiz': self.bufsiz,
            'input_bufsiz': self.input_bufsiz,
            'ranqd1': self.ranqd1,
            'stdin_in_use_by': self.stdin_in_use_by.decode() if self.stdin_in_use_by else None,
            'stdout_in_use_by': self.stdout_in_use_by.decode() if self.stdout_in_use_by else None,
            'subsystem': self.subsystem.decode() if self.subsystem else None,
            'tmp_path': self.tmp_path.decode() if self.tmp_path else None,
            'use_magic': self.use_magic,
            'use_threads': self.use_threads,
            'log2_dft_min_size': self.log2_dft_min_size
        }


    @property
    def verbosity(self) -> int:
        """messages are only written if globals.verbosity >= message.level"""
        return self.ptr.verbosity

    @verbosity.setter
    def verbosity(self, unsigned value):
        self.ptr.verbosity = value

    @property
    def repeatable(self) -> bool:
        """true to use pre-determined timestamps and PRNG seed"""
        return self.ptr.repeatable

    @repeatable.setter
    def repeatable(self, sox_bool value):
        self.ptr.repeatable = value

    @property
    def bufsiz(self) -> int:
        """Default size (in bytes) used by libSoX for blocks of sample data."""
        return self.ptr.bufsiz

    @bufsiz.setter
    def bufsiz(self, size_t value):
        self.ptr.bufsiz = value

    @property
    def input_bufsiz(self) -> int:
        """Default size (in bytes) used by libSoX for blocks of input sample data."""
        return self.ptr.input_bufsiz

    @input_bufsiz.setter
    def input_bufsiz(self, size_t value):
        self.ptr.input_bufsiz = value

    @property
    def ranqd1(self) -> int:
        """Can be used to re-seed libSoX's PRNG"""
        return self.ptr.ranqd1

    @ranqd1.setter
    def ranqd1(self, sox_int32_t value):
        self.ptr.ranqd1 = value

    @property
    def stdin_in_use_by(self) -> str:
        """Private: tracks the name of the handler currently using stdin"""
        if self.ptr.stdin_in_use_by == NULL:
            return None
        return self.ptr.stdin_in_use_by.decode()

    @property
    def stdout_in_use_by(self) -> str:
        """Private: tracks the name of the handler currently using stdout"""
        if self.ptr.stdout_in_use_by == NULL:
            return None
        return self.ptr.stdout_in_use_by.decode()

    @property
    def subsystem(self) -> str:
        """Private: tracks the name of the handler currently writing an output message"""
        if self.ptr.subsystem == NULL:
            return None
        return self.ptr.subsystem.decode()

    @property
    def tmp_path(self) -> str:
        """Private: client-configured path to use for temporary files"""
        if self.ptr.tmp_path == NULL:
            return None
        return self.ptr.tmp_path.decode()

    @property
    def use_magic(self) -> bool:
        """Private: true if client has requested use of 'magic' file-type detection"""
        return self.ptr.use_magic

    @property
    def use_threads(self) -> bool:
        """Private: true if client has requested parallel effects processing"""
        return self.ptr.use_threads

    @property
    def log2_dft_min_size(self) -> int:
        """Log to base 2 of minimum size (in bytes) used by libSoX for DFT (filtering)."""
        return self.ptr.log2_dft_min_size


cdef class EffectsGlobals:
    cdef sox_effects_globals_t* ptr
    cdef bint owner

    def __cinit__(self):
        self.ptr = NULL
        self.owner = False

    def __dealloc__(self):
        if self.ptr is not NULL and self.owner is True:
            free(self.ptr)
            self.ptr = NULL

    def __init__(self):
        self.ptr = sox_get_effects_globals()
        self.owner = False # never owned

    @staticmethod
    cdef EffectsGlobals from_ptr(sox_effects_globals_t* ptr, bint owner=False):
        cdef EffectsGlobals wrapper = EffectsGlobals.__new__(EffectsGlobals)
        wrapper.ptr = ptr
        wrapper.owner = owner
        return wrapper

    @property
    def plot(self) -> sox_plot_t:
        """To help the user choose effect & options"""
        return self.ptr.plot

    @plot.setter
    def plot(self, sox_plot_t value):
        self.ptr.plot = value

    @property
    def global_info(self) -> Globals:
        """Pointer to associated SoX globals"""
        if self.ptr.global_info == NULL:
            return None
        return Globals.from_ptr(self.ptr.global_info, False)


cdef class Format:
    """Audio format handle for reading and writing audio files.

    This class provides the main interface for opening, reading, and writing
    audio files in various formats supported by libsox.

    The Format class implements the context manager protocol, allowing automatic
    resource cleanup:

        with sox.Format('input.wav') as f:
            samples = f.read(1024)

    Attributes:
        filename: Path to the audio file
        signal: SignalInfo object describing audio properties
        encoding: EncodingInfo object describing encoding
        filetype: String identifier for the file format
        seekable: Whether seeking is supported
        mode: 'r' for reading, 'w' for writing
        olength: Samples * channels written to file
        clips: Counter incremented if clipping occurs
        sox_errno: Failure error code
        sox_errstr: Failure error text

    Example - Reading:
        >>> with sox.Format('input.wav') as f:
        ...     print(f"Rate: {f.signal.rate}, Channels: {f.signal.channels}")
        ...     samples = f.read(1024)

    Example - Writing:
        >>> signal = sox.SignalInfo(rate=44100, channels=2, precision=16)
        >>> with sox.Format('output.wav', signal=signal, mode='w') as f:
        ...     f.write([100, 200, 300, 400])
    """
    cdef sox_format_t* ptr
    cdef bint owner

    def __cinit__(self):
        self.ptr = NULL
        self.owner = False

    def __dealloc__(self):
        if self.ptr is not NULL and self.owner is True:
            sox_close(self.ptr)
            self.ptr = NULL

    def __init__(self, str filename, SignalInfo signal = None,
                 EncodingInfo encoding = None, str filetype = None, mode: str = 'r'):
        """Opens a file for reading or writing.

        Args:
            filename: Path to the audio file
            signal: Signal information (required for writing)
            encoding: Encoding information (optional)
            filetype: File format type (e.g., 'wav', 'coreaudio', 'alsa').
                      If None, format is inferred from filename extension.
            mode: 'r' for reading, 'w' for writing

        Raises:
            SoxFormatError: If file cannot be opened
            ValueError: If mode is invalid or signal missing for write mode
        """
        cdef char* filetype_c = NULL
        cdef bytes filetype_bytes
        if filetype is not None:
            filetype_bytes = filetype.encode('utf-8')
            filetype_c = filetype_bytes

        if mode == 'r':
            self.ptr = sox_open_read(
                filename.encode(),
                signal.ptr if signal else NULL,
                encoding.ptr if encoding else NULL,
                filetype_c
            )
        elif mode == 'w':
            if signal is None:
                raise ValueError("Signal information is required for writing")
            self.ptr = sox_open_write(
                filename.encode(),
                signal.ptr,
                encoding.ptr if encoding else NULL,
                filetype_c,
                NULL,
                NULL  # No overwrite callback for now
            )
        else:
            raise ValueError("Mode must be 'r' or 'w'")

        if self.ptr is NULL:
            raise SoxFormatError(f"Failed to open file {filename} in mode {mode}")

        self.owner = True

    @staticmethod
    cdef Format from_ptr(sox_format_t* ptr, bint owner=False):
        cdef Format wrapper = Format.__new__(Format)
        wrapper.ptr = ptr
        wrapper.owner = owner
        return wrapper

    def read(self, length: int) -> list:
        """Read samples from the file into a Python list.

        Args:
            length: Number of samples to read

        Returns:
            List of sample values (sox_sample_t integers)

        Note:
            For better performance with large data, use read_buffer() to get
            a memory view, or read_into() to read into a pre-allocated buffer.
        """
        cdef size_t samples_read = 0
        cdef sox_sample_t* buffer = <sox_sample_t*>malloc(length * sizeof(sox_sample_t))
        if buffer == NULL:
            raise SoxMemoryError("Failed to allocate read buffer")

        try:
            samples_read = sox_read(self.ptr, buffer, length)
            result = []
            for i in range(samples_read):
                result.append(buffer[i])
            return result
        finally:
            free(buffer)

    def read_buffer(self, length: int):
        """Read samples from the file into a buffer that supports the buffer protocol.

        Args:
            length: Number of samples to read

        Returns:
            A memory view of sox_sample_t values. Compatible with NumPy, PyTorch,
            array.array, and other buffer protocol objects.

        Example:
            >>> import numpy as np
            >>> with sox.Format('input.wav') as f:
            ...     buf = f.read_buffer(1024)
            ...     arr = np.asarray(buf, dtype=np.int32)
        """
        cdef size_t samples_read = 0
        cdef sox_sample_t* buffer = <sox_sample_t*>malloc(length * sizeof(sox_sample_t))
        if buffer == NULL:
            raise SoxMemoryError("Failed to allocate read buffer")

        samples_read = sox_read(self.ptr, buffer, length)

        # Create a memory view from the C buffer
        # We need to copy the data because we're freeing the buffer
        cdef sox_sample_t[::1] view = <sox_sample_t[:samples_read]>buffer
        result = bytearray(samples_read * sizeof(sox_sample_t))
        cdef unsigned char[::1] result_view = result
        cdef size_t i
        for i in range(samples_read * sizeof(sox_sample_t)):
            result_view[i] = (<unsigned char*>buffer)[i]

        free(buffer)
        return memoryview(result).cast('i', shape=[samples_read])

    def read_into(self, buffer) -> int:
        """Read samples from the file into a pre-allocated buffer.

        Args:
            buffer: Any object supporting the buffer protocol (numpy array,
                   array.array, bytearray, memoryview, etc.). Must be writable
                   and have enough space for sox_sample_t values.

        Returns:
            Number of samples actually read

        Example:
            >>> import numpy as np
            >>> arr = np.zeros(1024, dtype=np.int32)
            >>> with sox.Format('input.wav') as f:
            ...     n = f.read_into(arr)
            ...     print(f"Read {n} samples")
        """
        cdef Py_buffer pybuf
        cdef int flags = PyBUF_WRITABLE | PyBUF_C_CONTIGUOUS
        cdef size_t max_samples
        cdef size_t samples_read

        if PyObject_GetBuffer(buffer, &pybuf, flags) != 0:
            raise TypeError("Buffer must be writable and contiguous")

        try:
            # Calculate how many samples fit in the buffer
            max_samples = pybuf.len // sizeof(sox_sample_t)
            samples_read = sox_read(self.ptr, <sox_sample_t*>pybuf.buf, max_samples)
            return samples_read
        finally:
            PyBuffer_Release(&pybuf)

    def write(self, samples) -> int:
        """Write samples to the file.

        Args:
            samples: Can be a list, or any buffer protocol object (numpy array,
                    array.array, memoryview, etc.) containing sox_sample_t values

        Returns:
            Number of samples written

        Example:
            >>> import numpy as np
            >>> arr = np.array([100, 200, 300], dtype=np.int32)
            >>> with sox.Format('output.wav', signal=..., mode='w') as f:
            ...     f.write(arr)
        """
        cdef size_t length
        cdef sox_sample_t* buffer
        cdef size_t samples_written
        cdef Py_buffer pybuf
        cdef size_t i

        # Try buffer protocol first (more efficient)
        if PyObject_CheckBuffer(samples):
            if PyObject_GetBuffer(samples, &pybuf, PyBUF_C_CONTIGUOUS) != 0:
                raise TypeError("Buffer must be contiguous")

            try:
                length = pybuf.len // sizeof(sox_sample_t)
                samples_written = sox_write(self.ptr, <sox_sample_t*>pybuf.buf, length)
                return samples_written
            finally:
                PyBuffer_Release(&pybuf)

        # Fall back to list/sequence
        else:
            length = len(samples)
            buffer = <sox_sample_t*>malloc(length * sizeof(sox_sample_t))
            if buffer == NULL:
                raise SoxMemoryError("Failed to allocate write buffer")

            try:
                for i in range(length):
                    buffer[i] = samples[i]
                return <int>sox_write(self.ptr, buffer, length)
            finally:
                free(buffer)

    def seek(self, offset: int, whence: int = SOX_SEEK_SET) -> int:
        """Seek to a specific sample position."""
        cdef int result = sox_seek(self.ptr, offset, whence)
        if result != SOX_SUCCESS:
            raise SoxIOError(f"Failed to seek: {strerror(result)}")
        return result

    def close(self) -> int:
        """Closes an encoding or decoding session."""
        if self.ptr is NULL:
            # Already closed, return success
            return <int>SOX_SUCCESS

        cdef int result = sox_close(self.ptr)
        self.ptr = NULL  # Set to NULL regardless of result to prevent double-free
        if result != SOX_SUCCESS:
            raise SoxIOError(f"Failed to close: {strerror(result)}")
        return result

    def __enter__(self):
        """Context manager entry point.

        Returns:
            self: The Format object for use in with statement.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit point.

        Automatically closes the format when exiting the with block.

        Args:
            exc_type: Exception type if an exception occurred.
            exc_val: Exception value if an exception occurred.
            exc_tb: Exception traceback if an exception occurred.

        Returns:
            False: Allow exceptions to propagate.
        """
        if self.ptr is not NULL:
            try:
                self.close()
            except SoxIOError as e:
                # If close fails during exception handling, log but don't raise
                if exc_type is None:
                    raise
                else:
                    # Log the close error but don't mask the original exception
                    import sys
                    import warnings
                    warnings.warn(
                        f"Error closing format during exception handling: {e}",
                        ResourceWarning,
                        stacklevel=2
                    )
        return False

    @property
    def filename(self) -> str:
        """Path to the audio file."""
        return self.ptr.filename.decode()

    @property
    def signal(self) -> SignalInfo:
        """Signal information (sample rate, channels, precision, length)."""
        return SignalInfo.from_ptr(&self.ptr.signal)

    @property
    def encoding(self) -> EncodingInfo:
        """Encoding information (format, bits per sample, compression)."""
        return EncodingInfo.from_ptr(&self.ptr.encoding)

    @property
    def filetype(self) -> str:
        """File type identifier (e.g., 'wav', 'mp3', 'flac')."""
        if self.ptr.filetype == NULL:
            return None
        return self.ptr.filetype.decode()

    @property
    def seekable(self) -> bool:
        """Can seek on this file"""
        return self.ptr.seekable

    @property
    def mode(self) -> str:
        """Read or write mode ('r' or 'w')"""
        return chr(self.ptr.mode)

    @property
    def olength(self) -> int:
        """Samples * chans written to file"""
        return self.ptr.olength

    @property
    def clips(self) -> int:
        """Incremented if clipping occurs"""
        return self.ptr.clips

    @property
    def sox_errno(self) -> int:
        """Failure error code"""
        return self.ptr.sox_errno

    @property
    def sox_errstr(self) -> str:
        """Failure error text"""
        return self.ptr.sox_errstr.decode()

    @property
    def io_type(self) -> int:
        """Stores whether this is a file, pipe or URL"""
        return self.ptr.io_type

    @property
    def tell_off(self) -> int:
        """Current offset within file"""
        return self.ptr.tell_off

    @property
    def data_start(self) -> int:
        """Offset at which headers end and sound data begins"""
        return self.ptr.data_start


cdef class FormatHandler:
    """Handler structure defined by each format."""
    cdef sox_format_handler_t* ptr
    cdef bint owner

    def __cinit__(self):
        self.ptr = NULL
        self.owner = False

    def __dealloc__(self):
        if self.ptr is not NULL and self.owner is True:
            free(self.ptr)
            self.ptr = NULL

    def __init__(self, str path):
        self.ptr = <sox_format_handler_t*>sox_write_handler(
            path.encode(), NULL, NULL)
        self.owner = False  # sox_write_handler returns a static pointer, don't free it
        if self.ptr is NULL:
            raise SoxFormatError(f"Failed to find format handler for path: {path}")

    @staticmethod
    cdef FormatHandler from_ptr(sox_format_handler_t* ptr, bint owner=False):
        cdef FormatHandler wrapper = FormatHandler.__new__(FormatHandler)
        wrapper.ptr = ptr
        wrapper.owner = owner
        return wrapper

    @staticmethod
    def find(name: str, ignore_devices: bool = False) -> FormatHandler:
        """Finds a format handler by name.

        Note: Prefer using the module-level function `sox.find_format(name)`
        for better API consistency.
        """
        cdef bytes name_bytes = name.encode('utf-8')
        cdef const sox_format_handler_t* handler = sox_find_format(
            name_bytes, <sox_bool>ignore_devices)
        if handler == NULL:
            return None
        return FormatHandler.from_ptr(<sox_format_handler_t*>handler)

    @property
    def sox_lib_version_code(self) -> int:
        """Checked on load must be 1st in struct"""
        return self.ptr.sox_lib_version_code

    @property
    def description(self) -> str:
        """short description of format"""
        if self.ptr.description == NULL:
            return None
        return self.ptr.description.decode()

    @property
    def names(self) -> list:
        """null-terminated array of filename extensions that are handled by this format"""
        if self.ptr.names == NULL:
            return []

        result = []
        cdef int i = 0
        while self.ptr.names[i] != NULL:
            result.append(self.ptr.names[i].decode())
            i += 1
        return result

    @property
    def flags(self) -> int:
        """File flags (SOX_FILE_* values)."""
        return self.ptr.flags

    @property
    def priv_size(self) -> int:
        """Size of private data SoX should pre-allocate for format"""
        return self.ptr.priv_size


cdef class FormatTab:
    """Information about a loaded format handler, including the format name and a
    function pointer that can be invoked to get additional information about the
    format."""
    cdef sox_format_tab_t* ptr
    cdef bint owner

    def __cinit__(self):
        self.ptr = NULL
        self.owner = False

    def __dealloc__(self):
        if self.ptr is not NULL and self.owner is True:
            free(self.ptr)
            self.ptr = NULL

    def __init__(self, name: str = None):
        self.ptr = <sox_format_tab_t*>malloc(sizeof(sox_format_tab_t))
        if self.ptr is NULL:
            raise SoxMemoryError("Failed to allocate FormatTab")

        # Initialize struct members directly (name property is read-only)
        # Note: name parameter is ignored since setting it would be unsafe
        self.ptr.name = NULL
        self.ptr.fn = NULL
        self.owner = True

    @staticmethod
    cdef FormatTab from_ptr(sox_format_tab_t* ptr, bint owner=False):
        cdef FormatTab wrapper = FormatTab.__new__(FormatTab)
        wrapper.ptr = ptr
        wrapper.owner = owner
        return wrapper

    @property
    def name(self) -> str:
        """Name of format handler (read-only).

        Note: This property is read-only because the name pointer is managed
        by libsox and should not be modified externally.
        """
        if self.ptr.name == NULL:
            return None
        return self.ptr.name.decode()

    @property
    def fn(self):
        """Function to call to get format handler's information"""
        # Return None for function pointers as they can't be easily exposed to Python
        return None


cdef class EffectHandler:
    """Effect handler information."""
    cdef sox_effect_handler_t* ptr
    cdef bint owner

    def __cinit__(self):
        self.ptr = NULL
        self.owner = False

    def __dealloc__(self):
        if self.ptr is not NULL and self.owner is True:
            free(self.ptr)
            self.ptr = NULL

    @staticmethod
    cdef EffectHandler from_ptr(sox_effect_handler_t* ptr, bint owner=False):
        cdef EffectHandler wrapper = EffectHandler.__new__(EffectHandler)
        wrapper.ptr = ptr
        wrapper.owner = owner
        return wrapper

    @staticmethod
    def find(name: str) -> Effect:
        """Finds the effect handler with the given name.

        Note: Prefer using the module-level function `sox.find_effect(name)`
        for better API consistency.
        """
        cdef bytes name_bytes = name.encode('utf-8')
        cdef const sox_effect_handler_t* handler = sox_find_effect(name_bytes)
        if handler == NULL:
            return None
        return EffectHandler.from_ptr(<sox_effect_handler_t*>handler)

    @property
    def name(self) -> str:
        """Effect name"""
        if self.ptr.name == NULL:
            return None
        return self.ptr.name.decode()

    @property
    def usage(self) -> str:
        """Short explanation of parameters accepted by effect"""
        if self.ptr.usage == NULL:
            return None
        return self.ptr.usage.decode()

    @property
    def flags(self) -> int:
        """Combination of SOX_EFF_* flags"""
        return self.ptr.flags

    @property
    def priv_size(self) -> int:
        """Size of private data SoX should pre-allocate for effect"""
        return self.ptr.priv_size


cdef class Effect:
    """Effect structure."""
    cdef sox_effect_t* ptr
    cdef bint owner

    def __cinit__(self):
        self.ptr = NULL
        self.owner = False

    def __dealloc__(self):
        if self.ptr is not NULL and self.owner is True:
            free(self.ptr)
            self.ptr = NULL

    def __init__(self, handler: EffectHandler):
        """Initialize an effect with an optional handler."""
        self.ptr = sox_create_effect(handler.ptr)
        if self.ptr == NULL:
            raise SoxEffectError("Failed to create effect")
        self.owner = True

    @staticmethod
    cdef Effect from_ptr(sox_effect_t* ptr, bint owner=False):
        cdef Effect wrapper = Effect.__new__(Effect)
        wrapper.ptr = ptr
        wrapper.owner = owner
        return wrapper

    def delete(self):
        """Shut down and delete an effect."""
        sox_delete_effect(self.ptr)

    def set_options(self, options):
        """Applies the command-line options to the effect.

        Returns the number of arguments consumed."""
        cdef int argc = len(options) if options else 0
        cdef char** argv = <char**>malloc(argc * sizeof(char*)) if argc > 0 else NULL
        if argc > 0 and argv == NULL:
            raise SoxMemoryError("Failed to allocate argument array")

        cdef list byte_strings = []
        try:
            for i in range(argc):
                option = options[i]
                if isinstance(option, Format):
                    # For input/output effects, pass Format pointer cast to char*
                    argv[i] = <char*>(<Format>option).ptr
                elif isinstance(option, str):
                    # Encode string to bytes and keep reference
                    byte_str = option.encode('utf-8')
                    byte_strings.append(byte_str)
                    argv[i] = <char*>byte_str
                else:
                    raise TypeError(f"Unsupported option type: {type(option)}")
            return sox_effect_options(self.ptr, argc, argv)
        finally:
            if argv != NULL:
                free(argv)

    def stop(self) -> int:
        """Shuts down an effect (calls stop on each of its flows).

        Returns the number of clips from all flows."""
        if self.ptr == NULL:
            return 0
        return <int>sox_stop_effect(self.ptr)

    def trim_get_start(self) -> int:
        """Gets the sample offset of the start of the trim."""
        if self.ptr == NULL:
            return 0
        return <int>sox_trim_get_start(self.ptr)

    def trim_clear_start(self):
        """Clears the start of the trim to 0."""
        if self.ptr != NULL:
            sox_trim_clear_start(self.ptr)

    @property
    def global_info(self):
        """global effect parameters"""
        if self.ptr.global_info == NULL:
            return None
        return EffectsGlobals.from_ptr(self.ptr.global_info, False)

    @property
    def in_signal(self) -> SignalInfo:
        """Information about the incoming data stream"""
        return SignalInfo.from_ptr(&self.ptr.in_signal, False)

    @property
    def out_signal(self) -> SignalInfo:
        """Information about the outgoing data stream"""
        return SignalInfo.from_ptr(&self.ptr.out_signal, False)

    @property
    def in_encoding(self) -> EncodingInfo:
        """Information about the incoming data encoding"""
        if self.ptr.in_encoding == NULL:
            return None
        return EncodingInfo.from_ptr(self.ptr.in_encoding, False)

    @property
    def out_encoding(self) -> EncodingInfo:
        """Information about the outgoing data encoding"""
        if self.ptr.out_encoding == NULL:
            return None
        return EncodingInfo.from_ptr(self.ptr.out_encoding, False)

    @property
    def handler(self) -> EffectHandler:
        """The handler for this effect"""
        return EffectHandler.from_ptr(&self.ptr.handler, False)

    @property
    def clips(self) -> sox_uint64_t:
        """increment if clipping occurs"""
        return self.ptr.clips

    @property
    def flows(self) -> int:
        """1 if MCHAN, number of chans otherwise"""
        return self.ptr.flows

    @property
    def flow(self) -> int:
        """flow number"""
        return self.ptr.flow

    @property
    def priv(self):
        """Effect's private data area (each flow has a separate copy)"""
        # Return None for void* pointers as they can't be easily exposed to Python
        return None


# C callback function for flow_effects
# This is called by libsox and needs to invoke the Python callback
cdef int _flow_effects_callback_wrapper(sox_bool all_done, void* client_data) noexcept nogil:
    """C callback that wraps Python callback for flow_effects.

    This function is called from C code without the GIL, so we need to
    acquire it before calling Python code.
    """
    global _last_callback_exception
    with gil:
        try:
            # client_data is a pointer to the chain's address (as an integer)
            chain_id = <uintptr_t>client_data

            # Look up the callback in our global dictionary
            if chain_id not in _flow_callbacks:
                return SOX_SUCCESS  # No callback registered, continue

            callback, user_data = _flow_callbacks[chain_id]

            # Call the Python callback
            # all_done is a sox_bool (enum), convert to Python bool
            result = callback(bool(all_done), user_data)

            # If callback returns None or True, continue; False or int error code aborts
            if result is None or result is True:
                return SOX_SUCCESS
            elif result is False:
                return -1  # Generic error
            else:
                return int(result)  # Assume it's an error code

        except Exception as e:
            # Store the exception for later retrieval since we can't propagate through C code
            import sys
            _last_callback_exception = sys.exc_info()
            sys.stderr.write(f"Error in flow_effects callback: {e}\n")
            return -1


def get_last_callback_exception():
    """Retrieve the last exception that occurred in a flow_effects callback.

    Since exceptions cannot propagate through C code, they are stored for
    later retrieval. This function returns the exception info tuple
    (type, value, traceback) or None if no exception occurred.

    After calling this function, the stored exception is cleared.

    Returns:
        Tuple of (exception_type, exception_value, traceback) or None

    Example:
        >>> chain.flow_effects(callback=my_callback)
        >>> exc_info = sox.get_last_callback_exception()
        >>> if exc_info:
        ...     raise exc_info[1].with_traceback(exc_info[2])
    """
    global _last_callback_exception
    exc_info = _last_callback_exception
    _last_callback_exception = None
    return exc_info


cdef class EffectsChain:
    """Effects chain structure."""
    cdef sox_effects_chain_t* ptr
    cdef bint owner

    def __cinit__(self):
        self.ptr = NULL
        self.owner = False

    def __dealloc__(self):
        if self.ptr is not NULL and self.owner is True:
            sox_delete_effects_chain(self.ptr)
            self.ptr = NULL

    def __init__(self, in_encoding: EncodingInfo = None,
                       out_encoding: EncodingInfo = None):
        """Initialize an effects chain with optional input and output encodings."""
        self.ptr = sox_create_effects_chain(
            in_encoding.ptr if in_encoding else NULL,
            out_encoding.ptr if out_encoding else NULL)
        if self.ptr == NULL:
            raise SoxEffectError("Failed to create effects chain")
        self.owner = True

    @staticmethod
    cdef EffectsChain from_ptr(sox_effects_chain_t* ptr, bint owner=False):
        cdef EffectsChain wrapper = EffectsChain.__new__(EffectsChain)
        wrapper.ptr = ptr
        wrapper.owner = owner
        return wrapper

    def add_effect(self, effect: Effect, in_signal: SignalInfo,
                   out_signal: SignalInfo):
        """Adds an effect to the effects chain."""

        assert effect.ptr is not NULL, "effect is NULL"
        assert in_signal.ptr is not NULL, "in_signal is NULL"
        assert out_signal.ptr is not NULL, "out_signal is NULL"

        cdef int result = sox_add_effect(
            self.ptr, effect.ptr, in_signal.ptr, out_signal.ptr)
        if result != SOX_SUCCESS:
            raise SoxEffectError(f"Failed to add effect to chain: {strerror(result)}")
        return result

    def flow_effects(self, callback=None, client_data=None):
        """Runs the effects chain, optionally calling a callback for progress monitoring.

        Args:
            callback: Optional callable that receives (all_done: bool, user_data) and
                     returns None/True to continue, False to abort, or an integer error code.
                     Called periodically during processing (once per buffer).
            client_data: Optional user data passed to the callback

        Returns:
            SOX_SUCCESS if successful

        Raises:
            SoxEffectError: If the effects chain fails to run

        Example:
            >>> def progress_callback(all_done, user_data):
            ...     if all_done:
            ...         print("Processing complete!")
            ...     else:
            ...         print("Processing...")
            ...     return True  # Continue processing
            >>>
            >>> chain = sox.EffectsChain()
            >>> # ... add effects ...
            >>> chain.flow_effects(callback=progress_callback, client_data={'count': 0})
        """
        cdef int result
        cdef uintptr_t chain_id

        if callback is not None:
            # Store the callback in our global dictionary using chain pointer as key
            chain_id = <uintptr_t>self.ptr
            _flow_callbacks[chain_id] = (callback, client_data)

            try:
                # Call sox_flow_effects with our wrapper callback
                result = sox_flow_effects(
                    self.ptr,
                    _flow_effects_callback_wrapper,
                    <void*>chain_id
                )
            finally:
                # Clean up callback storage
                if chain_id in _flow_callbacks:
                    del _flow_callbacks[chain_id]
        else:
            # No callback, call with NULL
            result = sox_flow_effects(self.ptr, NULL, NULL)

        if result != SOX_SUCCESS:
            raise SoxEffectError(f"Failed to flow effects: {strerror(result)}")
        return result

    def get_clips(self) -> int:
        """Gets the number of clips that occurred while running the effects chain."""
        return <int>sox_effects_clips(self.ptr)

    def push_effect_last(self, effect: Effect):
        """Adds an already-initialized effect to the end of the chain."""
        if self.ptr != NULL and effect.ptr != NULL:
            sox_push_effect_last(self.ptr, effect.ptr)

    def pop_effect_last(self):
        """Removes and returns an effect from the end of the chain.

        Returns None if no effects."""
        cdef sox_effect_t* effect
        if self.ptr == NULL:
            return None
        effect = sox_pop_effect_last(self.ptr)
        if effect == NULL:
            return None
        return Effect.from_ptr(effect, True)

    def delete_effect_last(self):
        """Shut down and delete the last effect in the chain."""
        if self.ptr != NULL:
            sox_delete_effect_last(self.ptr)

    def delete_effects(self):
        """Shut down and delete all effects in the chain."""
        if self.ptr != NULL:
            sox_delete_effects(self.ptr)

    @property
    def effects(self) -> list:
        """Table of effects to be applied to a stream"""
        if self.ptr.effects == NULL:
            return []

        result = []
        for i in range(self.ptr.length):
            if self.ptr.effects[i] != NULL:
                result.append(Effect.from_ptr(self.ptr.effects[i], False))
        return result

    @property
    def length(self) -> int:
        """Number of effects to be applied"""
        return self.ptr.length

    @property
    def global_info(self) -> EffectsGlobals:
        """Copy of global effects settings"""
        return EffectsGlobals.from_ptr(&self.ptr.global_info, False)

    @property
    def in_enc(self) -> EncodingInfo:
        """Input encoding"""
        if self.ptr.in_enc == NULL:
            return None
        return EncodingInfo.from_ptr(self.ptr.in_enc, False)

    @property
    def out_enc(self) -> EncodingInfo:
        """Output encoding"""
        if self.ptr.out_enc == NULL:
            return None
        return EncodingInfo.from_ptr(self.ptr.out_enc, False)

    @property
    def table_size(self) -> int:
        """Size of effects table (including unused entries)"""
        return self.ptr.table_size

    @property
    def il_buf(self):
        """Channel interleave buffer"""
        # Return None for sox_sample_t* pointers as they can't be easily exposed to Python
        return None


# -------------------------------------------------------------------
# functions


def version() -> str:
    """Returns version number string of libSoX, for example, 14.4.0."""
    return (<const char*>sox_version()).decode()


def version_info():
    """Returns information about this build of libsox."""
    cdef const sox_version_info_t* info = sox_version_info()
    if info == NULL:
        return None

    return {
        'size': info.size,
        'flags': info.flags,
        'version_code': info.version_code,
        'version': info.version.decode() if info.version else None,
        'version_extra': info.version_extra.decode() if info.version_extra else None,
        'time': info.time.decode() if info.time else None,
        'distro': info.distro.decode() if info.distro else None,
        'compiler': info.compiler.decode() if info.compiler else None,
        'arch': info.arch.decode() if info.arch else None
    }





def get_encodings() -> list[EncodingsInfo]:
    """Returns the list of available encodings."""
    cdef const sox_encodings_info_t* encodings = sox_get_encodings_info()
    if encodings == NULL:
        return []

    result = []
    cdef int i = 0
    while encodings[i].name != NULL and encodings[i].flags < 3:
        result.append(
            EncodingsInfo(
                flags=encodings[i].flags,
                name=encodings[i].name.decode(),
                desc=encodings[i].desc.decode() if encodings[i].desc else None
            )
        )
        i += 1
    return result


def format_init():
    """Find and load format handler plugins.

    Raises:
        RuntimeError: If format initialization fails.
    """
    cdef int result = sox_format_init()
    if result != SOX_SUCCESS:
        raise SoxInitError(f"Failed to initialize format handlers: {strerror(result)}")


def format_quit():
    """Unload format handler plugins."""
    sox_format_quit()


def init():
    """Initialize the SoX effects library.

    Must be called before using any SoX functionality. This initializes
    the effects system and format handlers.

    This function is idempotent - calling it multiple times is safe.
    Subsequent calls after the first are no-ops.

    Raises:
        SoxInitError: If initialization fails.

    Example:
        >>> import cysox as sox
        >>> sox.init()
        >>> # Use sox functions...
        >>> sox.quit()
    """
    global _sox_initialized
    if _sox_initialized:
        return  # Already initialized, no-op
    cdef int result = sox_init()
    if result != SOX_SUCCESS:
        raise SoxInitError(f"Failed to initialize SoX library: {strerror(result)}")
    _sox_initialized = True


def quit():
    """Close the SoX effects library and unload format handlers.

    Note:
        Due to libsox limitations (crashes on re-initialization after quit),
        this function is a no-op during normal execution. Actual cleanup
        happens automatically at process exit via atexit. This ensures the
        library remains usable throughout the process lifetime.

        This function is retained for API compatibility but does not
        actually call sox_quit(). Use _force_quit() for testing if needed.

    Example:
        >>> import cysox as sox
        >>> sox.init()
        >>> # Use sox functions...
        >>> sox.quit()  # No-op; cleanup happens at process exit
    """
    # No-op: actual cleanup happens via atexit to prevent crashes from
    # repeated init/quit cycles. See KNOWN_LIMITATIONS.md
    pass


def _force_quit():
    """Internal: Actually quit sox (called by atexit handler).

    Warning: Do not call this directly - it will crash if sox is used again.
    This is only for use by the atexit cleanup handler.
    """
    global _sox_initialized
    if not _sox_initialized:
        return
    cdef int result = sox_quit()
    _sox_initialized = False
    if result != SOX_SUCCESS:
        raise SoxInitError(f"Failed to cleanup SoX library: {strerror(result)}")


def strerror(sox_errno: int) -> str:
    """Converts a SoX error code into an error string."""
    return (<const char*>sox_strerror(sox_errno)).decode()


def is_playlist(filename: str) -> bool:
    """Returns true if the specified file is a known playlist file type."""
    cdef bytes filename_bytes = filename.encode('utf-8')
    return sox_is_playlist(filename_bytes)


def basename(filename: str) -> str:
    """Gets the basename of the specified file.

    Args:
        filename: Path to extract basename from

    Returns:
        The basename portion of the filename
    """
    cdef bytes filename_bytes = filename.encode('utf-8')
    cdef size_t filename_len = len(filename_bytes)
    # Allocate buffer based on input length (basename can't be longer than full path)
    # Use at least 256 bytes for short paths, or filename length + 1 for longer paths
    cdef size_t buffer_size = filename_len + 1 if filename_len >= 256 else 256
    cdef char* base_buffer = <char*>malloc(buffer_size)

    if base_buffer is NULL:
        raise SoxMemoryError("Failed to allocate buffer for basename")

    try:
        result = sox_basename(base_buffer, buffer_size, filename_bytes)
        if result == 0:
            return ""
        return base_buffer[:result].decode('utf-8')
    finally:
        free(base_buffer)


def precision(sox_encoding_t encoding, unsigned bits_per_sample) -> int:
    """Given an encoding and bits_per_sample

    returns the number of useful bits per sample."""
    return sox_precision(encoding, bits_per_sample)


def get_effects_globals():
    """Returns global parameters for effects."""
    cdef sox_effects_globals_t* globals = sox_get_effects_globals()
    if globals == NULL:
        return None

    return {
        'plot': globals.plot
    }


def format_supports_encoding(str path, EncodingInfo encoding) -> bool:
    """Returns true if the format handler for the specified file type supports the specified encoding."""
    return bool(sox_format_supports_encoding(path.encode(), NULL, encoding.ptr))


def open_mem_read(buffer: bytes, signal: SignalInfo = None, encoding: EncodingInfo = None, filetype: str = None):
    """Opens a decoding session for a memory buffer. Returns a Format object or raises MemoryError."""
    cdef sox_format_t* fmt
    cdef char* filetype_c = NULL
    cdef bytes filetype_bytes
    if filetype is not None:
        filetype_bytes = filetype.encode('utf-8')
        filetype_c = filetype_bytes
    fmt = sox_open_mem_read(<void*>buffer, len(buffer),
                            signal.ptr if signal else NULL,
                            encoding.ptr if encoding else NULL,
                            filetype_c)
    if fmt == NULL:
        raise SoxFormatError("Failed to open memory buffer for reading")
    return Format.from_ptr(fmt, True)


def open_mem_write(buffer: bytearray, signal: SignalInfo,
        encoding: EncodingInfo, filetype: str = None, oob: OutOfBand = None):
    """Opens an encoding session for a memory buffer.

    Returns a Format object or raises MemoryError."""
    cdef sox_format_t* fmt
    cdef char* filetype_c = NULL
    cdef bytes filetype_bytes
    if filetype is not None:
        filetype_bytes = filetype.encode('utf-8')
        filetype_c = filetype_bytes
    fmt = sox_open_mem_write(<void*>buffer, len(buffer),
                             signal.ptr,
                             encoding.ptr,
                             filetype_c,
                             oob.ptr if oob else NULL)
    if fmt == NULL:
        raise SoxFormatError("Failed to open memory buffer for writing")
    return Format.from_ptr(fmt, True)


def open_memstream_write(signal: SignalInfo, encoding: EncodingInfo,
        filetype: str = None, oob: OutOfBand = None):
    """Opens an encoding session for a memstream buffer.

    Returns (Format, buffer_ptr, buffer_size_ptr) or raises MemoryError."""
    cdef sox_format_t* fmt
    cdef char* filetype_c = NULL
    cdef bytes filetype_bytes
    cdef char* buffer_ptr = NULL
    cdef size_t buffer_size = 0
    if filetype is not None:
        filetype_bytes = filetype.encode('utf-8')
        filetype_c = filetype_bytes
    fmt = sox_open_memstream_write(&buffer_ptr, &buffer_size,
                                   signal.ptr,
                                   encoding.ptr,
                                   filetype_c,
                                   oob.ptr if oob else NULL)
    if fmt == NULL:
        raise SoxFormatError("Failed to open memstream for writing")
    return Format.from_ptr(fmt, True), buffer_ptr, buffer_size


def get_format_fns():
    """Returns the table of format handler names and functions."""
    cdef const sox_format_tab_t* format_fns = sox_get_format_fns()
    if format_fns == NULL:
        return []

    result = []
    cdef int i = 0
    while format_fns[i].name != NULL:
        result.append({
            'name': format_fns[i].name.decode(),
            'fn': None  # Function pointers can't be easily exposed to Python
        })
        i += 1
    return result


def get_effect_fns():
    """Returns an array containing the known effect handlers."""
    cdef const sox_effect_fn_t* effect_fns = sox_get_effect_fns()
    if effect_fns == NULL:
        return []

    result = []
    cdef int i = 0
    while effect_fns[i] != NULL:
        # Note: We can't easily call the function pointer from Python
        # This is mainly for enumeration purposes
        result.append(i)
        i += 1
    return result


# cdef int playlist_callback(void * callback, char * filename) noexcept:
#     return SOX_SUCCESS


# def parse_playlist(callback, listname: str):
#     """Parses the specified playlist file."""
#     cdef int result
#     cdef bytes listname_bytes = listname.encode()

#     result = sox_parse_playlist(playlist_callback, NULL, listname_bytes)
#     if result != SOX_SUCCESS:
#         raise RuntimeError(f"Failed to parse playlist: {strerror(result)}")
#     return result


def read_samples(format: Format, buffer: list, length: int) -> int:
    """Reads samples from a decoding session into a sample buffer."""
    cdef size_t samples_read = 0
    cdef sox_sample_t* sample_buffer = <sox_sample_t*>malloc(length * sizeof(sox_sample_t))
    if sample_buffer == NULL:
        raise SoxMemoryError("Failed to allocate sample buffer")

    try:
        samples_read = sox_read(format.ptr, sample_buffer, length)
        # Convert C samples to Python list
        for i in range(samples_read):
            buffer.append(sample_buffer[i])
        return <int>samples_read
    finally:
        free(sample_buffer)


def write_samples(format: Format, samples: list) -> int:
    """Writes samples to an encoding session from a sample buffer."""
    cdef size_t length = len(samples)
    cdef sox_sample_t* sample_buffer = <sox_sample_t*>malloc(length * sizeof(sox_sample_t))
    if sample_buffer == NULL:
        raise SoxMemoryError("Failed to allocate sample buffer")

    try:
        # Convert Python list to C samples
        for i in range(length):
            sample_buffer[i] = samples[i]
        return <int>sox_write(format.ptr, sample_buffer, length)
    finally:
        free(sample_buffer)


def seek_samples(format: Format, offset: int, whence: int = SOX_SEEK_SET) -> int:
    """Sets the location at which next samples will be decoded."""
    cdef int result = sox_seek(format.ptr, offset, whence)
    if result != SOX_SUCCESS:
        raise SoxIOError(f"Failed to seek: {strerror(result)}")
    return result


def find_effect(name: str) -> EffectHandler:
    """Finds the effect handler with the given name.

    Args:
        name: Name of the effect (e.g., 'trim', 'rate', 'vol', 'reverb')

    Returns:
        EffectHandler object if found, None otherwise

    Example:
        >>> handler = sox.find_effect('trim')
        >>> if handler:
        ...     effect = sox.Effect(handler)
        ...     effect.set_options(['0', '10'])
    """
    cdef bytes name_bytes = name.encode('utf-8')
    cdef const sox_effect_handler_t* handler = sox_find_effect(name_bytes)
    if handler == NULL:
        return None
    return EffectHandler.from_ptr(<sox_effect_handler_t*>handler)


def find_format(name: str, ignore_devices: bool = False) -> FormatHandler:
    """Finds a format handler by name."""
    cdef bytes name_bytes = name.encode('utf-8')
    cdef const sox_format_handler_t* handler = sox_find_format(name_bytes, <sox_bool>ignore_devices)
    if handler == NULL:
        return None
    return FormatHandler.from_ptr(<sox_format_handler_t*>handler)

# def flow_effects(EffectsChain chain):
#     assert sox_flow_effects(chain.ptr, NULL, NULL) == SOX_SUCCESS




"""Filter effects (highpass, lowpass, bandpass, etc.)."""

from typing import List

from .base import Effect


class HighPass(Effect):
    """High-pass filter (removes low frequencies).

    Args:
        frequency: Cutoff frequency in Hz.
        poles: Filter order, 1 or 2 (default: 2). Higher = steeper rolloff.

    Example:
        >>> fx.HighPass(frequency=80)     # Remove rumble below 80Hz
        >>> fx.HighPass(frequency=200, poles=1)  # Gentle rolloff
    """

    def __init__(self, frequency: float, *, poles: int = 2):
        if poles not in (1, 2):
            raise ValueError("poles must be 1 or 2")
        self.frequency = frequency
        self.poles = poles

    @property
    def name(self) -> str:
        return "highpass"

    def to_args(self) -> List[str]:
        return [f"-{self.poles}", str(self.frequency)]


class LowPass(Effect):
    """Low-pass filter (removes high frequencies).

    Args:
        frequency: Cutoff frequency in Hz.
        poles: Filter order, 1 or 2 (default: 2). Higher = steeper rolloff.

    Example:
        >>> fx.LowPass(frequency=8000)    # Remove highs above 8kHz
        >>> fx.LowPass(frequency=4000, poles=1)  # Gentle rolloff
    """

    def __init__(self, frequency: float, *, poles: int = 2):
        if poles not in (1, 2):
            raise ValueError("poles must be 1 or 2")
        self.frequency = frequency
        self.poles = poles

    @property
    def name(self) -> str:
        return "lowpass"

    def to_args(self) -> List[str]:
        return [f"-{self.poles}", str(self.frequency)]


class BandPass(Effect):
    """Band-pass filter (passes frequencies within a range).

    Args:
        frequency: Center frequency in Hz.
        width: Filter width. Interpretation depends on width_type.
        width_type: 'q' for Q-factor, 'h' for Hz, 'o' for octaves (default: 'q').
        constant_skirt: Use constant skirt gain (default: False).

    Example:
        >>> fx.BandPass(frequency=1000, width=2)  # Q=2 bandpass at 1kHz
    """

    def __init__(
        self,
        frequency: float,
        width: float,
        *,
        width_type: str = "q",
        constant_skirt: bool = False
    ):
        if width_type not in ("q", "h", "o"):
            raise ValueError("width_type must be 'q', 'h', or 'o'")
        self.frequency = frequency
        self.width = width
        self.width_type = width_type
        self.constant_skirt = constant_skirt

    @property
    def name(self) -> str:
        return "bandpass"

    def to_args(self) -> List[str]:
        args = []
        if self.constant_skirt:
            args.append("-c")
        args.append(str(self.frequency))
        args.append(f"{self.width}{self.width_type}")
        return args


class BandReject(Effect):
    """Band-reject (notch) filter.

    Args:
        frequency: Center frequency in Hz.
        width: Filter width. Interpretation depends on width_type.
        width_type: 'q' for Q-factor, 'h' for Hz, 'o' for octaves (default: 'q').

    Example:
        >>> fx.BandReject(frequency=60, width=10)  # Remove 60Hz hum
    """

    def __init__(
        self,
        frequency: float,
        width: float,
        *,
        width_type: str = "q"
    ):
        if width_type not in ("q", "h", "o"):
            raise ValueError("width_type must be 'q', 'h', or 'o'")
        self.frequency = frequency
        self.width = width
        self.width_type = width_type

    @property
    def name(self) -> str:
        return "bandreject"

    def to_args(self) -> List[str]:
        return [str(self.frequency), f"{self.width}{self.width_type}"]



"""Tests for buffer protocol support in Format class"""
import pytest
import array
from cysox import sox


@pytest.fixture(autouse=True)
def initialize_sox():
    """Initialize SoX before each test."""
    sox.init()
    yield
    sox.quit()


def test_read_buffer():
    """Test Format.read_buffer() returns a memoryview"""
    with sox.Format('tests/data/s00.wav') as f:
        buf = f.read_buffer(100)
        assert buf is not None
        assert isinstance(buf, memoryview)
        assert len(buf) <= 100


def test_read_into_array():
    """Test Format.read_into() with array.array"""
    arr = array.array('i', [0] * 100)  # signed 32-bit int array

    with sox.Format('tests/data/s00.wav') as f:
        n = f.read_into(arr)
        assert n > 0
        assert n <= 100
        # First sample should be non-zero (actual audio data)
        # Commented out as actual value depends on audio file content
        # assert arr[0] != 0


def test_write_from_list():
    """Test Format.write() with list (backward compatibility)"""
    import tempfile
    import os

    signal = sox.SignalInfo(rate=44100, channels=1, precision=16)

    with tempfile.TemporaryDirectory() as tmpdir:
        output_path = os.path.join(tmpdir, 'output.wav')

        with sox.Format(output_path, signal=signal, mode='w') as f:
            samples = [1000, 2000, 3000, 4000]
            n = f.write(samples)
            assert n == len(samples)


def test_write_from_array():
    """Test Format.write() with array.array (buffer protocol)"""
    import tempfile
    import os

    signal = sox.SignalInfo(rate=44100, channels=1, precision=16)

    with tempfile.TemporaryDirectory() as tmpdir:
        output_path = os.path.join(tmpdir, 'output.wav')

        with sox.Format(output_path, signal=signal, mode='w') as f:
            arr = array.array('i', [1000, 2000, 3000, 4000])
            n = f.write(arr)
            assert n == len(arr)


def test_buffer_roundtrip():
    """Test reading and writing using buffer protocol"""
    import tempfile
    import os

    with tempfile.TemporaryDirectory() as tmpdir:
        output_path = os.path.join(tmpdir, 'roundtrip.wav')

        # Read some samples
        with sox.Format('tests/data/s00.wav') as f:
            buf = f.read_buffer(100)
            signal = sox.SignalInfo(
                rate=f.signal.rate,
                channels=f.signal.channels,
                precision=f.signal.precision
            )

        # Write them back
        with sox.Format(output_path, signal=signal, mode='w') as f:
            # Convert memoryview to array for writing
            arr = array.array('i', buf)
            n = f.write(arr)
            assert n == len(buf)

        # Read them back again and verify
        with sox.Format(output_path) as f:
            buf2 = f.read_buffer(100)
            assert len(buf2) == len(buf)



from cysox import sox


# Test Globals class


def test_globals():
    """Test get_globals function"""
    g = sox.Globals()
    assert isinstance(g, sox.Globals)
    assert g.verbosity == 2
    assert g.repeatable == 0
    assert g.bufsiz == 8192



from cysox import utils


def test_utls_lib_version():
    assert utils.lib_version(14, 4, 2) == 918530


def test_utls_lib_version_code():
    assert utils.lib_version_code() == 918530


def test_utils_int_max():
    assert utils.int_max(8) == 127


def test_utils_int_min():
    assert utils.int_min(2) == 2


def test_utils_uint_max():
    assert utils.uint_max(10) == 1023



from cysox import sox


# Test OutOfBand class
# def test_out_of_band_creation():
#     """Test OutOfBand creation"""
#     oob = sox.OutOfBand()
#     assert oob is not None


# def test_out_of_band_comments():
#     """Test OutOfBand comment methods"""
#     oob = sox.OutOfBand()

#     # Test initial state
#     assert oob.num_comments() == 0
#     assert oob.comments == []

#     # Test appending comments
#     oob.append_comment("artist=Test Artist")
#     assert oob.num_comments() == 1
#     assert "artist=Test Artist" in oob.comments

#     oob.append_comment("title=Test Title")
#     assert oob.num_comments() == 2
#     assert "artist=Test Artist" in oob.comments
#     assert "title=Test Title" in oob.comments

#     # Test appending multiple comments
#     oob.append_comments("album=Test Album\nyear=2023")
#     assert oob.num_comments() == 4
#     assert "album=Test Album" in oob.comments
#     assert "year=2023" in oob.comments


# def test_out_of_band_find_comment():
#     """Test OutOfBand find_comment method"""
#     oob = sox.OutOfBand()
#     oob.append_comment("artist=Test Artist")
#     oob.append_comment("title=Test Title")

#     # Test finding existing comment
#     assert oob.find_comment("artist") == "Test Artist"
#     assert oob.find_comment("title") == "Test Title"

#     # Test finding non-existent comment
#     assert oob.find_comment("nonexistent") is None


# def test_out_of_band_instr():
#     """Test OutOfBand instr property"""
#     oob = sox.OutOfBand()
#     instr = oob.instr

#     assert isinstance(instr, sox.InstrInfo)
#     assert instr.note == 0
#     assert instr.low == 0
#     assert instr.high == 0
#     assert instr.loopmode == 0
#     assert instr.nloops == 0


# def test_out_of_band_memory_management():
#     """Test OutOfBand memory management"""
#     oob = sox.OutOfBand()
#     oob.append_comment("test=value")
#     assert oob.num_comments() == 1

#     # Test that the object can be properly cleaned up
#     del oob



# Type stubs for cysox.utils

def lib_version(major: int, minor: int, patch: int) -> int:
    """Compute a 32-bit integer API version from three 8-bit parts."""
    ...

def lib_version_code() -> int:
    """Get the current libSoX version code."""
    ...

def int_min(bits: int) -> int:
    """Returns the smallest (negative) value storable in a twos-complement signed integer."""
    ...

def int_max(bits: int) -> int:
    """Returns the largest (positive) value storable in a twos-complement signed integer."""
    ...

def uint_max(bits: int) -> int:
    """Returns the largest value storable in an unsigned integer."""
    ...



from cysox import sox


# Test FileInfo class
def test_file_info_creation():
    """Test FileInfo creation and properties

    Note: buf parameter is ignored for safety (read-only property).
    FileInfo is primarily for internal use by libsox.
    """
    test_data = b"test data"
    file_info = sox.FileInfo(buf=test_data, size=len(test_data), count=5, pos=2)

    # buf is read-only and always None when created via __init__ (for safety)
    assert file_info.buf is None
    assert file_info.size == len(test_data)
    assert file_info.count == 5
    assert file_info.pos == 2


def test_file_info_default_values():
    """Test FileInfo with default values"""
    file_info = sox.FileInfo()

    assert file_info.buf is None
    assert file_info.size == 0
    assert file_info.count == 0
    assert file_info.pos == 0


def test_file_info_property_setters():
    """Test FileInfo property setters (except buf which is read-only)"""
    file_info = sox.FileInfo()

    # buf is read-only for safety (no setter)
    # We can only set the other properties
    file_info.size = 13
    file_info.count = 10
    file_info.pos = 5

    assert file_info.buf is None  # Always None, can't be set
    assert file_info.size == 13
    assert file_info.count == 10
    assert file_info.pos == 5


# def test_file_info_buf_none():
#     """Test FileInfo buf property when set to None"""
#     test_data = b"test data"
#     file_info = sox.FileInfo(buf=test_data)
#     assert file_info.buf == test_data

#     file_info.buf = None
#     assert file_info.buf is None


# def test_file_info_memory_management():
#     """Test FileInfo memory management"""
#     file_info = sox.FileInfo(buf=b"test", size=4, count=1, pos=0)
#     assert file_info.size == 4

#     # Test that the object can be properly cleaned up
#     del file_info



"""Tests for example1: Apply vol & flanger effects
Port of tests/examples/example1.c
"""
import pytest
import tempfile
import os
from cysox import sox


@pytest.fixture(autouse=True)
def initialize_sox():
    """Initialize SoX before each test."""
    sox.init()
    yield
    sox.quit()


def test_example1_vol_and_flanger():
    """Test applying vol and flanger effects to an audio file.

    This is a port of example1.c which reads an input file, applies
    vol (3dB) and flanger effects, and writes to an output file.
    """
    input_file = 'tests/data/s00.wav'

    with tempfile.TemporaryDirectory() as tmpdir:
        output_file = os.path.join(tmpdir, 'output.wav')

        # Open input file
        in_fmt = sox.Format(input_file)

        # Open output file with same characteristics as input
        signal = sox.SignalInfo(
            rate=in_fmt.signal.rate,
            channels=in_fmt.signal.channels,
            precision=in_fmt.signal.precision
        )
        out_fmt = sox.Format(output_file, signal=signal, mode='w')

        # Create effects chain
        chain = sox.EffectsChain(in_fmt.encoding, out_fmt.encoding)

        # Add input effect
        input_handler = sox.find_effect("input")
        input_effect = sox.Effect(input_handler)
        input_effect.set_options([in_fmt])
        chain.add_effect(input_effect, in_fmt.signal, in_fmt.signal)

        # Add vol effect with 3dB gain
        vol_handler = sox.find_effect("vol")
        vol_effect = sox.Effect(vol_handler)
        vol_effect.set_options(["3dB"])
        chain.add_effect(vol_effect, in_fmt.signal, in_fmt.signal)

        # Add flanger effect with default parameters
        flanger_handler = sox.find_effect("flanger")
        flanger_effect = sox.Effect(flanger_handler)
        flanger_effect.set_options([])
        chain.add_effect(flanger_effect, in_fmt.signal, in_fmt.signal)

        # Add output effect
        output_handler = sox.find_effect("output")
        output_effect = sox.Effect(output_handler)
        output_effect.set_options([out_fmt])
        chain.add_effect(output_effect, in_fmt.signal, in_fmt.signal)

        # Flow samples through the effects chain
        result = chain.flow_effects()
        assert result == sox.SUCCESS

        # Cleanup
        in_fmt.close()
        out_fmt.close()

        # Verify output file was created and has content
        assert os.path.exists(output_file)
        assert os.path.getsize(output_file) > 0

        # Verify we can read the output file
        verify_fmt = sox.Format(output_file)
        assert verify_fmt.signal.rate == signal.rate
        assert verify_fmt.signal.channels == signal.channels
        verify_fmt.close()



"""Tests for example6: Explicit format conversion
Port of tests/examples/example6.c
"""
import pytest
import tempfile
import os
from cysox import sox


@pytest.fixture(autouse=True)
def initialize_sox():
    """Initialize SoX before each test."""
    sox.init()
    yield
    sox.quit()


def test_example6_explicit_conversion():
    """Test explicit format conversion to mono mu-law at 8kHz.

    This is a port of example6.c which demonstrates explicitly specifying
    output file signal and encoding attributes to convert any input file
    to mono mu-law at 8kHz sampling rate.
    """
    input_file = 'tests/data/s00.wav'

    with tempfile.TemporaryDirectory() as tmpdir:
        output_file = os.path.join(tmpdir, 'output.wav')

        # Open input file
        in_fmt = sox.Format(input_file)

        # Specify output encoding: mu-law, 8-bit
        # SOX_ENCODING_ULAW = 9 (from sox.pxd enum sox_encoding_t)
        out_encoding = sox.EncodingInfo(
            encoding=9,  # ULAW encoding
            bits_per_sample=8
        )

        # Specify output signal: 8kHz, mono
        out_signal = sox.SignalInfo(
            rate=8000,
            channels=1
        )

        # Open output file with explicit signal and encoding
        out_fmt = sox.Format(output_file, signal=out_signal,
                            encoding=out_encoding, mode='w')

        # Create effects chain
        chain = sox.EffectsChain(in_fmt.encoding, out_fmt.encoding)

        # Intermediate signal (starts as input characteristics)
        interm_signal = sox.SignalInfo(
            rate=in_fmt.signal.rate,
            channels=in_fmt.signal.channels,
            precision=in_fmt.signal.precision
        )

        # Add input effect
        input_handler = sox.find_effect("input")
        input_effect = sox.Effect(input_handler)
        input_effect.set_options([in_fmt])
        chain.add_effect(input_effect, interm_signal, in_fmt.signal)

        # Add rate conversion if needed
        if in_fmt.signal.rate != out_fmt.signal.rate:
            rate_handler = sox.find_effect("rate")
            rate_effect = sox.Effect(rate_handler)
            rate_effect.set_options([])  # Use default parameters
            chain.add_effect(rate_effect, interm_signal, out_fmt.signal)

        # Add channel conversion if needed
        if in_fmt.signal.channels != out_fmt.signal.channels:
            channels_handler = sox.find_effect("channels")
            channels_effect = sox.Effect(channels_handler)
            channels_effect.set_options([])  # Use default parameters
            chain.add_effect(channels_effect, interm_signal, out_fmt.signal)

        # Add output effect
        output_handler = sox.find_effect("output")
        output_effect = sox.Effect(output_handler)
        output_effect.set_options([out_fmt])
        chain.add_effect(output_effect, interm_signal, out_fmt.signal)

        # Flow samples through the chain
        result = chain.flow_effects()
        assert result == sox.SUCCESS

        # Cleanup
        in_fmt.close()
        out_fmt.close()

        # Verify output file was created
        assert os.path.exists(output_file)
        assert os.path.getsize(output_file) > 0

        # Verify output characteristics
        verify_fmt = sox.Format(output_file)
        assert verify_fmt.signal.rate == 8000, \
            f"Expected rate 8000, got {verify_fmt.signal.rate}"
        assert verify_fmt.signal.channels == 1, \
            f"Expected 1 channel, got {verify_fmt.signal.channels}"
        assert verify_fmt.encoding.encoding == 9, \
            f"Expected ULAW encoding (9), got {verify_fmt.encoding.encoding}"
        assert verify_fmt.encoding.bits_per_sample == 8, \
            f"Expected 8 bits per sample, got {verify_fmt.encoding.bits_per_sample}"
        verify_fmt.close()


def test_example6_different_output_format():
    """Test conversion to different output format specifications."""
    input_file = 'tests/data/s00.wav'

    with tempfile.TemporaryDirectory() as tmpdir:
        output_file = os.path.join(tmpdir, 'output.wav')

        in_fmt = sox.Format(input_file)

        # Test with different output specs: 16kHz, stereo, signed PCM
        # SOX_ENCODING_SIGN2 = 1 (from sox.pxd enum sox_encoding_t)
        out_encoding = sox.EncodingInfo(
            encoding=1,  # SIGN2 encoding
            bits_per_sample=16
        )

        out_signal = sox.SignalInfo(
            rate=16000,
            channels=2
        )

        out_fmt = sox.Format(output_file, signal=out_signal,
                            encoding=out_encoding, mode='w')

        chain = sox.EffectsChain(in_fmt.encoding, out_fmt.encoding)

        interm_signal = sox.SignalInfo(
            rate=in_fmt.signal.rate,
            channels=in_fmt.signal.channels,
            precision=in_fmt.signal.precision
        )

        # Input effect
        input_handler = sox.find_effect("input")
        input_effect = sox.Effect(input_handler)
        input_effect.set_options([in_fmt])
        chain.add_effect(input_effect, interm_signal, in_fmt.signal)

        # Rate conversion
        if in_fmt.signal.rate != out_fmt.signal.rate:
            rate_handler = sox.find_effect("rate")
            rate_effect = sox.Effect(rate_handler)
            rate_effect.set_options([])
            chain.add_effect(rate_effect, interm_signal, out_fmt.signal)

        # Channel conversion
        if in_fmt.signal.channels != out_fmt.signal.channels:
            channels_handler = sox.find_effect("channels")
            channels_effect = sox.Effect(channels_handler)
            channels_effect.set_options([])
            chain.add_effect(channels_effect, interm_signal, out_fmt.signal)

        # Output effect
        output_handler = sox.find_effect("output")
        output_effect = sox.Effect(output_handler)
        output_effect.set_options([out_fmt])
        chain.add_effect(output_effect, interm_signal, out_fmt.signal)

        result = chain.flow_effects()
        assert result == sox.SUCCESS

        in_fmt.close()
        out_fmt.close()

        # Verify output characteristics
        verify_fmt = sox.Format(output_file)
        assert verify_fmt.signal.rate == 16000
        assert verify_fmt.signal.channels == 2
        assert verify_fmt.encoding.bits_per_sample == 16
        verify_fmt.close()



import math

from cysox import sox


def test_get_encodings():
    """Test get_encodings function"""
    encodings = sox.get_encodings()
    assert isinstance(encodings, list)
    assert len(encodings) > 0

    for encoding in encodings:
        assert hasattr(encoding, "flags")
        assert hasattr(encoding, "name")
        assert hasattr(encoding, "desc")
        assert hasattr(encoding, "type")


def test_encodings():
    """Test that ENCODINGS list is properly defined"""
    assert hasattr(sox, "ENCODINGS")
    assert isinstance(sox.ENCODINGS, list)
    assert len(sox.ENCODINGS) > 0

    # Test some known encodings
    assert ("SIGN2", "signed linear 2's comp: Mac") in sox.ENCODINGS
    assert ("MP3", "MP3 compression") in sox.ENCODINGS
    assert ("FLOAT", "floating point (binary format)") in sox.ENCODINGS


# Test EncodingInfo class
def test_encoding_info_creation():
    """Test EncodingInfo creation and properties"""
    encoding = sox.EncodingInfo(
        encoding=1,  # SIGN2
        bits_per_sample=16,
        compression=1.0,
        reverse_bytes=0,
        reverse_nibbles=0,
        reverse_bits=0,
        opposite_endian=False,
    )

    assert encoding.encoding == 1
    assert encoding.bits_per_sample == 16
    assert encoding.compression == 1.0
    assert encoding.reverse_bytes == 0
    assert encoding.reverse_nibbles == 0
    assert encoding.reverse_bits == 0
    assert not encoding.opposite_endian


def test_encoding_info_default_values():
    """Test EncodingInfo with default values"""
    encoding = sox.EncodingInfo()

    assert encoding.encoding == 0
    assert encoding.bits_per_sample == 0
    assert encoding.compression == 0.0
    assert encoding.reverse_bytes == 0
    assert encoding.reverse_nibbles == 0
    assert encoding.reverse_bits == 0
    assert not encoding.opposite_endian


def test_encoding_info_property_setters():
    """Test EncodingInfo property setters"""
    encoding = sox.EncodingInfo()

    encoding.encoding = 22  # MP3
    encoding.bits_per_sample = 0
    encoding.compression = math.inf
    encoding.reverse_bytes = 1
    encoding.reverse_nibbles = 1
    encoding.reverse_bits = 1
    encoding.opposite_endian = True

    assert encoding.encoding == 22
    assert encoding.bits_per_sample == 0
    assert encoding.compression == math.inf
    assert encoding.reverse_bytes == 1
    assert encoding.reverse_nibbles == 1
    assert encoding.reverse_bits == 1
    assert encoding.opposite_endian


def test_encoding_info_property_access():
    """Test EncodingInfo property access patterns"""
    encoding = sox.EncodingInfo(encoding=1, bits_per_sample=16)

    # Test getting properties multiple times
    assert encoding.encoding == 1
    assert encoding.encoding == 1  # Should be consistent

    assert encoding.bits_per_sample == 16
    assert encoding.bits_per_sample == 16  # Should be consistent

    # Test setting and getting
    encoding.encoding = 22  # MP3
    assert encoding.encoding == 22

    encoding.bits_per_sample = 0
    assert encoding.bits_per_sample == 0


def test_encoding_info_edge_cases():
    """Test EncodingInfo with edge case values"""
    encoding = sox.EncodingInfo(
        encoding=25,  # High encoding value
        bits_per_sample=32,  # High bits per sample
        compression=100.0,  # High compression
        reverse_bytes=1,
        reverse_nibbles=1,
        reverse_bits=1,
        opposite_endian=True,
    )

    assert encoding.encoding == 25
    assert encoding.bits_per_sample == 32
    assert encoding.compression == 100.0
    assert encoding.reverse_bytes == 1
    assert encoding.reverse_nibbles == 1
    assert encoding.reverse_bits == 1
    assert encoding.opposite_endian


def test_encoding_info_memory_management():
    """Test EncodingInfo memory management"""
    encoding = sox.EncodingInfo(encoding=1, bits_per_sample=16)
    assert encoding.encoding == 1

    # Test that the object can be properly cleaned up
    del encoding



"""Tests for example4: Concatenate audio files
Port of tests/examples/example4.c
"""
import pytest
import tempfile
import os
from cysox import sox


@pytest.fixture(autouse=True)
def initialize_sox():
    """Initialize SoX before each test."""
    sox.init()
    yield
    sox.quit()


def test_example4_concatenate_files():
    """Test concatenating multiple audio files into one.

    This is a port of example4.c which concatenates audio files.
    Files must have the same number of channels and sample rate.
    """
    # For testing, we'll use the same file twice to demonstrate concatenation
    input_file = 'tests/data/s00.wav'
    max_samples = 2048

    with tempfile.TemporaryDirectory() as tmpdir:
        output_file = os.path.join(tmpdir, 'concatenated.wav')

        # We'll concatenate the same input file twice
        input_files = [input_file, input_file]

        output = None
        signal = None

        try:
            for i, input_path in enumerate(input_files):
                # Open input file
                input_fmt = sox.Format(input_path)

                if i == 0:
                    # First file: open output with same characteristics
                    signal = sox.SignalInfo(
                        rate=input_fmt.signal.rate,
                        channels=input_fmt.signal.channels,
                        precision=input_fmt.signal.precision
                    )
                    # Note: encoding.encoding is a sox_encoding_t enum value (int)
                    encoding = sox.EncodingInfo(
                        encoding=int(input_fmt.encoding.encoding),
                        bits_per_sample=input_fmt.encoding.bits_per_sample
                    )
                    output = sox.Format(output_file, signal=signal, encoding=encoding, mode='w')
                else:
                    # Subsequent files: check signal matches
                    assert input_fmt.signal.channels == signal.channels, \
                        "All input files must have same number of channels"
                    assert input_fmt.signal.rate == signal.rate, \
                        "All input files must have same sample rate"

                # Copy all samples from this input to output
                while True:
                    samples = input_fmt.read(max_samples)
                    if len(samples) == 0:
                        break  # EOF

                    written = output.write(samples)
                    assert written == len(samples), \
                        f"Write failed: wrote {written} of {len(samples)} samples"

                input_fmt.close()

            # Close output
            output.close()
            output = None

            # Verify output file exists and has content
            assert os.path.exists(output_file)
            assert os.path.getsize(output_file) > 0

            # Verify we can read the output file and it has expected characteristics
            verify_fmt = sox.Format(output_file)
            assert verify_fmt.signal.rate == signal.rate
            assert verify_fmt.signal.channels == signal.channels

            # The output should have approximately twice the samples of one input
            # (allowing for header overhead)
            input_fmt = sox.Format(input_file)
            input_length = input_fmt.signal.length
            output_length = verify_fmt.signal.length

            # Output should be roughly 2x input length (within 10% tolerance)
            if input_length > 0 and output_length > 0:
                expected = input_length * len(input_files)
                ratio = output_length / expected
                assert 0.9 <= ratio <= 1.1, \
                    f"Output length {output_length} not close to 2x input {input_length}"

            input_fmt.close()
            verify_fmt.close()

        except Exception as e:
            # Cleanup on error
            if output is not None:
                output.close()
            if os.path.exists(output_file):
                os.remove(output_file)
            raise


def test_example4_mismatched_channels():
    """Test that concatenating files with different channels fails."""
    input_file = 'tests/data/s00.wav'

    with tempfile.TemporaryDirectory() as tmpdir:
        # Create a mono version of the input for testing
        mono_file = os.path.join(tmpdir, 'mono.wav')
        output_file = os.path.join(tmpdir, 'output.wav')

        # First create a mono file
        in_fmt = sox.Format(input_file)
        mono_signal = sox.SignalInfo(
            rate=in_fmt.signal.rate,
            channels=1,  # Force mono
            precision=in_fmt.signal.precision
        )

        # Use effects chain to convert to mono
        out_fmt = sox.Format(mono_file, signal=mono_signal, mode='w')
        chain = sox.EffectsChain(in_fmt.encoding, out_fmt.encoding)

        interm_signal = sox.SignalInfo(
            rate=in_fmt.signal.rate,
            channels=in_fmt.signal.channels,
            precision=in_fmt.signal.precision
        )

        input_handler = sox.find_effect("input")
        input_effect = sox.Effect(input_handler)
        input_effect.set_options([in_fmt])
        chain.add_effect(input_effect, interm_signal, in_fmt.signal)

        if in_fmt.signal.channels != mono_signal.channels:
            channels_handler = sox.find_effect("channels")
            channels_effect = sox.Effect(channels_handler)
            channels_effect.set_options([])
            chain.add_effect(channels_effect, interm_signal, mono_signal)

        output_handler = sox.find_effect("output")
        output_effect = sox.Effect(output_handler)
        output_effect.set_options([out_fmt])
        chain.add_effect(output_effect, interm_signal, mono_signal)

        chain.flow_effects()
        in_fmt.close()
        out_fmt.close()

        # Now try to concatenate mono and stereo files (should fail)
        if os.path.exists(mono_file):
            first_fmt = sox.Format(input_file)
            second_fmt = sox.Format(mono_file)

            # Check that they have different channel counts
            channels_match = first_fmt.signal.channels == second_fmt.signal.channels

            first_fmt.close()
            second_fmt.close()

            if not channels_match:
                # This is the expected behavior - files have different channels
                # In the actual concatenation code, this would raise an assertion
                pass



"""Equalization effects (bass, treble, equalizer)."""

from typing import List

from .base import Effect


class Bass(Effect):
    """Boost or cut bass frequencies.

    Args:
        gain: Amount to boost (positive) or cut (negative) in dB.
        frequency: Center frequency in Hz (default: 100).
        width: Filter width in octaves (default: 0.5).

    Example:
        >>> fx.Bass(gain=5)              # Boost bass by 5dB
        >>> fx.Bass(gain=-3, frequency=80)  # Cut at 80Hz
    """

    def __init__(
        self,
        gain: float,
        *,
        frequency: float = 100,
        width: float = 0.5
    ):
        self.gain = gain
        self.frequency = frequency
        self.width = width

    @property
    def name(self) -> str:
        return "bass"

    def to_args(self) -> List[str]:
        return [str(self.gain), str(self.frequency), str(self.width)]


class Treble(Effect):
    """Boost or cut treble frequencies.

    Args:
        gain: Amount to boost (positive) or cut (negative) in dB.
        frequency: Center frequency in Hz (default: 3000).
        width: Filter width in octaves (default: 0.5).

    Example:
        >>> fx.Treble(gain=3)             # Boost treble by 3dB
        >>> fx.Treble(gain=-2, frequency=4000)  # Cut at 4kHz
    """

    def __init__(
        self,
        gain: float,
        *,
        frequency: float = 3000,
        width: float = 0.5
    ):
        self.gain = gain
        self.frequency = frequency
        self.width = width

    @property
    def name(self) -> str:
        return "treble"

    def to_args(self) -> List[str]:
        return [str(self.gain), str(self.frequency), str(self.width)]


class Equalizer(Effect):
    """Peaking EQ filter at a specific frequency.

    Args:
        frequency: Center frequency in Hz.
        width: Filter width (Q factor or bandwidth in Hz).
        gain: Amount to boost (positive) or cut (negative) in dB.

    Example:
        >>> fx.Equalizer(frequency=1000, width=1, gain=3)  # Boost 1kHz
    """

    def __init__(self, frequency: float, width: float, gain: float):
        self.frequency = frequency
        self.width = width
        self.gain = gain

    @property
    def name(self) -> str:
        return "equalizer"

    def to_args(self) -> List[str]:
        return [str(self.frequency), str(self.width), str(self.gain)]



"""Base classes for typed audio effects."""

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    import numpy as np


class Effect(ABC):
    """Base class for all typed effects.

    Subclasses must implement:
    - name: The sox effect name
    - to_args(): Convert parameters to sox argument list
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """The sox effect name."""
        pass

    @abstractmethod
    def to_args(self) -> List[str]:
        """Convert effect parameters to sox argument list."""
        pass

    def _repr_args(self) -> str:
        """Return string representation of arguments for __repr__."""
        args = []
        for key, value in self.__dict__.items():
            if not key.startswith('_'):
                args.append(f"{key}={value!r}")
        return ", ".join(args)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._repr_args()})"


class CompositeEffect(Effect):
    """Base class for effects that combine multiple sox effects.

    Subclasses must implement the `effects` property returning a list
    of Effect instances to be applied in sequence.

    Example:
        class WarmReverb(CompositeEffect):
            def __init__(self, decay=60):
                self.decay = decay

            @property
            def effects(self):
                return [
                    HighPass(frequency=80),
                    Reverb(reverberance=self.decay),
                    Volume(db=-2),
                ]
    """

    @property
    @abstractmethod
    def effects(self) -> List[Effect]:
        """Return list of constituent effects."""
        pass

    @property
    def name(self) -> str:
        return self.__class__.__name__

    def to_args(self) -> List[str]:
        raise TypeError(
            f"CompositeEffect '{self.name}' must be expanded, not converted to args. "
            "Use expand_effects() to get the list of constituent effects."
        )


class PythonEffect(Effect):
    """Base class for custom Python-based sample processing.

    Subclasses must implement the `process()` method which receives
    samples as a numpy array and returns processed samples.

    Note: Python effects require numpy and run outside the sox pipeline,
    making them slower than native sox effects. Use for custom DSP that
    sox doesn't support.

    Example:
        class BitCrusher(PythonEffect):
            def __init__(self, bits=8):
                self.bits = bits

            def process(self, samples, sample_rate, channels):
                levels = 2 ** self.bits
                return np.round(samples * levels) / levels
    """

    @property
    def name(self) -> str:
        return self.__class__.__name__

    def to_args(self) -> List[str]:
        raise TypeError(
            f"PythonEffect '{self.name}' cannot be converted to sox args. "
            "It will be processed separately using the process() method."
        )

    @abstractmethod
    def process(
        self,
        samples: "np.ndarray",
        sample_rate: int,
        channels: int
    ) -> "np.ndarray":
        """Process audio samples.

        Args:
            samples: Input samples as numpy array of shape (n_samples,) for mono
                    or (n_samples, channels) for multi-channel.
            sample_rate: Sample rate in Hz.
            channels: Number of audio channels.

        Returns:
            Processed samples as numpy array with same shape as input.
        """
        pass


class CEffect(Effect):
    """Base class for custom C-level effects.

    For advanced users who implement effects in C or Cython and need
    to register them with sox.

    Subclasses should:
    1. Set _handler_ptr to the pointer returned by their Cython module
    2. Call register() once at startup before using the effect
    """

    _handler_ptr: int = None

    @classmethod
    def register(cls) -> None:
        """Register this effect's handler with sox.

        Must be called once before using the effect.
        """
        if cls._handler_ptr is None:
            raise ValueError(f"{cls.__name__}._handler_ptr is not set")

        from cysox import sox
        if hasattr(sox, 'register_effect_handler'):
            sox.register_effect_handler(cls._handler_ptr)
        else:
            raise NotImplementedError(
                "Custom C effect registration not yet implemented in low-level API"
            )

    @property
    @abstractmethod
    def name(self) -> str:
        """The registered effect name."""
        pass

    @abstractmethod
    def to_args(self) -> List[str]:
        """Convert parameters to sox argument list."""
        pass



"""Tests for example5: Memory-based I/O
Port of tests/examples/example5.c
"""
import pytest
import tempfile
import os
from cysox import sox


@pytest.fixture(autouse=True)
def initialize_sox():
    """Initialize SoX before each test."""
    sox.init()
    yield
    sox.quit()


@pytest.mark.skip(reason="Memory I/O (memstream) crashes during write - libsox memory I/O may not be fully functional")
def test_example5_memory_io():
    """Test reading and writing audio files in memory buffers.

    This is a port of example5.c which demonstrates using memory buffers
    instead of files for audio I/O. We read a file, write to memory,
    then read from memory and write to a file.

    NOTE: Even using sox_open_memstream_write (which allocates its own C buffer),
    the code still crashes during write operations. This suggests libsox's memory I/O
    implementation may not be fully functional or has platform-specific issues.

    Attempted fixes:
    - Fixed filetype string encoding (was corrupting strings)
    - Switched from bytearray to memstream (C-managed buffer)
    - Still crashes at write time
    """
    input_file = 'tests/data/s00.wav'
    max_samples = 2048

    with tempfile.TemporaryDirectory() as tmpdir:
        output_file = os.path.join(tmpdir, 'output.wav')

        # Step 1: Read from file and write to memory buffer
        # Use memstream which allocates its own buffer
        in_fmt = sox.Format(input_file)

        # Create memory-based output using open_memstream_write
        # This allocates its own buffer that won't be moved by Python GC
        out_mem, buffer_ptr, buffer_size = sox.open_memstream_write(
            in_fmt.signal,
            in_fmt.encoding,
            filetype="wav"
        )

        # Copy samples from file to memory
        total_written = 0
        while True:
            samples = in_fmt.read(max_samples)
            if len(samples) == 0:
                break  # EOF

            written = out_mem.write(samples)
            assert written == len(samples), "Memory write failed"
            total_written += written

        in_fmt.close()
        out_mem.close()

        assert total_written > 0, "Should have written some samples to memory"

        # Step 2: Read from memory buffer and write to file
        # Copy the C buffer to Python bytes
        # Note: buffer_ptr is a char* and buffer_size is size_t
        import ctypes
        memory_buffer = ctypes.string_at(buffer_ptr, buffer_size)

        # Open memory buffer for reading
        in_mem = sox.open_mem_read(memory_buffer, filetype="wav")

        # Create output file
        out_signal = sox.SignalInfo(
            rate=in_mem.signal.rate,
            channels=in_mem.signal.channels,
            precision=in_mem.signal.precision
        )
        out_fmt = sox.Format(output_file, signal=out_signal, mode='w')

        # Copy samples from memory to file
        total_read = 0
        while True:
            samples = in_mem.read(max_samples)
            if len(samples) == 0:
                break  # EOF

            written = out_fmt.write(samples)
            assert written == len(samples), "File write failed"
            total_read += written

        in_mem.close()
        out_fmt.close()

        assert total_read > 0, "Should have read some samples from memory"

        # Verify output file exists and has content
        assert os.path.exists(output_file)
        assert os.path.getsize(output_file) > 0

        # Verify we can read the output file
        verify_fmt = sox.Format(output_file)
        assert verify_fmt.signal.rate == out_signal.rate
        assert verify_fmt.signal.channels == out_signal.channels
        verify_fmt.close()


@pytest.mark.skip(reason="Memory I/O (memstream) crashes during write - libsox memory I/O may not be fully functional")
def test_example5_roundtrip_samples():
    """Test that samples survive a memory I/O roundtrip."""
    import ctypes
    input_file = 'tests/data/s00.wav'
    num_samples = 1000  # Read a specific number of samples

    # Read original samples
    in_fmt = sox.Format(input_file)
    original_samples = in_fmt.read(num_samples)

    # Write to memory using memstream
    out_mem, buffer_ptr, buffer_size = sox.open_memstream_write(
        in_fmt.signal,
        in_fmt.encoding,
        filetype="wav"
    )
    out_mem.write(original_samples)
    out_mem.close()
    in_fmt.close()

    # Copy C buffer to Python bytes
    memory_buffer = ctypes.string_at(buffer_ptr, buffer_size)

    # Read back from memory
    in_mem = sox.open_mem_read(memory_buffer, filetype="wav")
    readback_samples = in_mem.read(num_samples)
    in_mem.close()

    # Verify we got the same number of samples back
    assert len(readback_samples) == len(original_samples), \
        f"Expected {len(original_samples)} samples, got {len(readback_samples)}"

    # Note: We don't compare the actual sample values because WAV format
    # encoding/decoding may introduce small differences. The important
    # thing is that the roundtrip works and we get the right number of samples.



"""Format conversion effects (rate, channels, bits)."""

from typing import List, Optional

from .base import Effect


class Rate(Effect):
    """Resample to a different sample rate.

    Args:
        sample_rate: Target sample rate in Hz.
        quality: Resampling quality - 'quick', 'low', 'medium', 'high',
                or 'very-high' (default: 'high').

    Example:
        >>> fx.Rate(sample_rate=48000)
        >>> fx.Rate(sample_rate=44100, quality='very-high')
    """

    QUALITY_FLAGS = {
        "quick": "-q",
        "low": "-l",
        "medium": "-m",
        "high": "-h",
        "very-high": "-v",
    }

    def __init__(self, sample_rate: int, *, quality: str = "high"):
        if quality not in self.QUALITY_FLAGS:
            raise ValueError(
                f"quality must be one of: {', '.join(self.QUALITY_FLAGS.keys())}"
            )
        self.sample_rate = sample_rate
        self.quality = quality

    @property
    def name(self) -> str:
        return "rate"

    def to_args(self) -> List[str]:
        return [self.QUALITY_FLAGS[self.quality], str(self.sample_rate)]


class Channels(Effect):
    """Change the number of audio channels.

    Args:
        channels: Target number of channels (1=mono, 2=stereo, etc.).

    Example:
        >>> fx.Channels(channels=1)   # Convert to mono
        >>> fx.Channels(channels=2)   # Convert to stereo
    """

    def __init__(self, channels: int):
        if channels < 1:
            raise ValueError("channels must be at least 1")
        self.channels = channels

    @property
    def name(self) -> str:
        return "channels"

    def to_args(self) -> List[str]:
        return [str(self.channels)]


class Remix(Effect):
    """Remix channels with custom mix specification.

    Args:
        mix: Channel mix specification. Each output channel is specified
             as a string like "1,2" (mix channels 1 and 2) or "1v0.5,2v0.5"
             (mix with volume adjustment). Use "-" for silence.

    Example:
        >>> fx.Remix(mix=["1,2"])           # Mono mixdown
        >>> fx.Remix(mix=["1", "2"])        # Keep stereo as-is
        >>> fx.Remix(mix=["2", "1"])        # Swap L/R
        >>> fx.Remix(mix=["1v0.5,2v0.5"])   # Mono with equal mix
    """

    def __init__(self, mix: List[str]):
        self.mix = mix

    @property
    def name(self) -> str:
        return "remix"

    def to_args(self) -> List[str]:
        return self.mix


class Dither(Effect):
    """Apply dithering for bit-depth reduction.

    Args:
        type: Dither type - 'rectangular', 'triangular', 'gaussian',
              or 'shaped' (default: 'shaped').
        precision: Target precision in bits (optional).

    Example:
        >>> fx.Dither()                      # Default shaped dither
        >>> fx.Dither(type='triangular')     # TPDF dither
    """

    def __init__(
        self,
        *,
        type: str = "shaped",
        precision: Optional[int] = None
    ):
        valid_types = ("rectangular", "triangular", "gaussian", "shaped")
        if type not in valid_types:
            raise ValueError(f"type must be one of: {', '.join(valid_types)}")
        self.type = type
        self.precision = precision

    @property
    def name(self) -> str:
        return "dither"

    def to_args(self) -> List[str]:
        args = [f"-{self.type[0]}"]  # First letter
        if self.precision is not None:
            args.extend(["-p", str(self.precision)])
        return args



"""Tests for example3: Playback with trim and format conversion
Port of tests/examples/example3.c
"""
import pytest
import tempfile
import os
from cysox import sox


@pytest.fixture(autouse=True)
def initialize_sox():
    """Initialize SoX before each test."""
    sox.init()
    yield
    sox.quit()


def test_example3_trim_and_conversion():
    """Test trimming audio and handling rate/channel conversion.

    This is a port of example3.c which plays an audio file starting
    at 10 seconds with sample-rate and channel conversion if needed.
    For testing, we write to a file instead of playing to audio device.
    """
    input_file = 'tests/data/s00.wav'

    with tempfile.TemporaryDirectory() as tmpdir:
        output_file = os.path.join(tmpdir, 'output.wav')

        # Open input file
        in_fmt = sox.Format(input_file)

        # For testing, we'll create output with different rate/channels
        # to exercise the conversion logic
        out_signal = sox.SignalInfo(
            rate=22050,  # Different from input (typically 44100)
            channels=1,  # Mono (input might be stereo)
            precision=in_fmt.signal.precision
        )
        out_fmt = sox.Format(output_file, signal=out_signal, mode='w')

        # Create effects chain
        chain = sox.EffectsChain(in_fmt.encoding, out_fmt.encoding)

        # Store intermediate signal
        interm_signal = sox.SignalInfo(
            rate=in_fmt.signal.rate,
            channels=in_fmt.signal.channels,
            precision=in_fmt.signal.precision
        )

        # Add input effect
        input_handler = sox.find_effect("input")
        input_effect = sox.Effect(input_handler)
        input_effect.set_options([in_fmt])
        chain.add_effect(input_effect, interm_signal, in_fmt.signal)

        # Add trim effect to start at a specific position
        # Using 0.1 seconds for testing (original uses 10 seconds)
        trim_handler = sox.find_effect("trim")
        trim_effect = sox.Effect(trim_handler)
        trim_effect.set_options(["0.1"])
        chain.add_effect(trim_effect, interm_signal, in_fmt.signal)

        # Add rate conversion if needed
        if in_fmt.signal.rate != out_fmt.signal.rate:
            rate_handler = sox.find_effect("rate")
            rate_effect = sox.Effect(rate_handler)
            rate_effect.set_options([])
            chain.add_effect(rate_effect, interm_signal, out_fmt.signal)

        # Add channel conversion if needed
        if in_fmt.signal.channels != out_fmt.signal.channels:
            channels_handler = sox.find_effect("channels")
            channels_effect = sox.Effect(channels_handler)
            channels_effect.set_options([])
            chain.add_effect(channels_effect, interm_signal, out_fmt.signal)

        # Add output effect
        output_handler = sox.find_effect("output")
        output_effect = sox.Effect(output_handler)
        output_effect.set_options([out_fmt])
        chain.add_effect(output_effect, interm_signal, out_fmt.signal)

        # Flow samples through the chain
        result = chain.flow_effects()
        assert result == sox.SUCCESS

        # Cleanup
        in_fmt.close()
        out_fmt.close()

        # Verify output file was created
        assert os.path.exists(output_file)
        assert os.path.getsize(output_file) > 0

        # Verify output characteristics
        verify_fmt = sox.Format(output_file)
        assert verify_fmt.signal.rate == 22050
        assert verify_fmt.signal.channels == 1
        verify_fmt.close()


def test_example3_no_conversion():
    """Test trim effect without rate/channel conversion."""
    input_file = 'tests/data/s00.wav'

    with tempfile.TemporaryDirectory() as tmpdir:
        output_file = os.path.join(tmpdir, 'output.wav')

        in_fmt = sox.Format(input_file)

        # Output with same characteristics as input (no conversion needed)
        out_signal = sox.SignalInfo(
            rate=in_fmt.signal.rate,
            channels=in_fmt.signal.channels,
            precision=in_fmt.signal.precision
        )
        out_fmt = sox.Format(output_file, signal=out_signal, mode='w')

        chain = sox.EffectsChain(in_fmt.encoding, out_fmt.encoding)

        interm_signal = sox.SignalInfo(
            rate=in_fmt.signal.rate,
            channels=in_fmt.signal.channels,
            precision=in_fmt.signal.precision
        )

        # Input effect
        input_handler = sox.find_effect("input")
        input_effect = sox.Effect(input_handler)
        input_effect.set_options([in_fmt])
        chain.add_effect(input_effect, interm_signal, in_fmt.signal)

        # Trim effect
        trim_handler = sox.find_effect("trim")
        trim_effect = sox.Effect(trim_handler)
        trim_effect.set_options(["0.1"])
        chain.add_effect(trim_effect, interm_signal, in_fmt.signal)

        # Output effect (no rate/channel conversion needed)
        output_handler = sox.find_effect("output")
        output_effect = sox.Effect(output_handler)
        output_effect.set_options([out_fmt])
        chain.add_effect(output_effect, interm_signal, out_fmt.signal)

        result = chain.flow_effects()
        assert result == sox.SUCCESS

        in_fmt.close()
        out_fmt.close()

        # Verify output
        assert os.path.exists(output_file)
        assert os.path.getsize(output_file) > 0



"""Tests for the high-level cysox API."""

import pytest
import cysox
from cysox import fx


class TestInfo:
    """Tests for cysox.info()."""

    def test_info_returns_dict(self, test_wav_str):
        """info() returns a dictionary."""
        info = cysox.info(test_wav_str)
        assert isinstance(info, dict)

    def test_info_contains_required_keys(self, test_wav_str):
        """info() returns all required keys."""
        info = cysox.info(test_wav_str)
        required_keys = [
            "path",
            "format",
            "duration",
            "sample_rate",
            "channels",
            "bits_per_sample",
            "samples",
            "encoding",
        ]
        for key in required_keys:
            assert key in info, f"Missing key: {key}"

    def test_info_values(self, test_wav_str):
        """info() returns correct values for test file."""
        info = cysox.info(test_wav_str)
        assert info["path"] == test_wav_str
        assert info["format"] == "wav"
        assert info["sample_rate"] == 44100
        assert info["channels"] == 2
        assert info["bits_per_sample"] == 16
        assert info["samples"] > 0
        assert info["duration"] > 0
        assert info["encoding"] == "signed-integer"

    def test_info_nonexistent_file(self):
        """info() raises exception for nonexistent file."""
        with pytest.raises(Exception):
            cysox.info("/nonexistent/file.wav")


class TestConvert:
    """Tests for cysox.convert()."""

    def test_simple_conversion(self, test_wav_str, output_path):
        """Basic file conversion without effects."""
        cysox.convert(test_wav_str, output_path)
        assert output_path.exists()

        info = cysox.info(str(output_path))
        original = cysox.info(test_wav_str)
        assert info["sample_rate"] == original["sample_rate"]
        assert info["channels"] == original["channels"]

    def test_conversion_with_effects(self, test_wav_str, output_path):
        """Conversion with effects chain."""
        cysox.convert(
            test_wav_str,
            output_path,
            effects=[
                fx.Volume(db=3),
                fx.Bass(gain=2),
            ]
        )
        assert output_path.exists()
        assert output_path.stat().st_size > 0

    def test_conversion_with_sample_rate(self, test_wav_str, output_path):
        """Conversion with sample rate change."""
        cysox.convert(test_wav_str, output_path, sample_rate=22050)

        info = cysox.info(str(output_path))
        assert info["sample_rate"] == 22050

    def test_conversion_with_channels(self, test_wav_str, output_path):
        """Conversion with channel count change."""
        cysox.convert(test_wav_str, output_path, channels=1)

        info = cysox.info(str(output_path))
        assert info["channels"] == 1

    def test_conversion_with_all_options(self, test_wav_str, output_path):
        """Conversion with effects and format options."""
        cysox.convert(
            test_wav_str,
            output_path,
            effects=[fx.Normalize(), fx.Fade(fade_in=0.1)],
            sample_rate=22050,
            channels=1,
        )

        info = cysox.info(str(output_path))
        assert info["sample_rate"] == 22050
        assert info["channels"] == 1


class TestStream:
    """Tests for cysox.stream()."""

    def test_stream_yields_memoryview(self, test_wav_str):
        """stream() yields memoryview objects."""
        for chunk in cysox.stream(test_wav_str, chunk_size=1024):
            assert isinstance(chunk, memoryview)
            break  # Just test first chunk

    def test_stream_reads_all_samples(self, test_wav_str):
        """stream() reads all samples from file."""
        info = cysox.info(test_wav_str)
        expected_samples = info["samples"]

        total_samples = 0
        for chunk in cysox.stream(test_wav_str, chunk_size=8192):
            # memoryview len() returns number of int32 elements (samples)
            total_samples += len(chunk)

        assert total_samples == expected_samples

    def test_stream_chunk_size(self, test_wav_str):
        """stream() respects chunk_size parameter."""
        chunk_size = 1000
        for chunk in cysox.stream(test_wav_str, chunk_size=chunk_size):
            # Last chunk may be smaller
            assert len(chunk) // 4 <= chunk_size
            break

    def test_stream_empty_on_nonexistent_file(self):
        """stream() raises exception for nonexistent file."""
        with pytest.raises(Exception):
            list(cysox.stream("/nonexistent/file.wav"))


class TestFxClasses:
    """Tests for fx effect classes."""

    def test_volume_effect(self):
        """Volume effect creates correct args."""
        v = fx.Volume(db=3)
        assert v.name == "vol"
        assert v.to_args() == ["3dB"]

        v = fx.Volume(db=-6, limiter=True)
        assert v.to_args() == ["-6dB", "limiter"]

    def test_bass_effect(self):
        """Bass effect creates correct args."""
        b = fx.Bass(gain=5)
        assert b.name == "bass"
        assert b.to_args() == ["5", "100", "0.5"]

    def test_treble_effect(self):
        """Treble effect creates correct args."""
        t = fx.Treble(gain=-2, frequency=4000)
        assert t.name == "treble"
        assert t.to_args() == ["-2", "4000", "0.5"]

    def test_reverb_effect(self):
        """Reverb effect creates correct args."""
        r = fx.Reverb(reverberance=80)
        assert r.name == "reverb"
        args = r.to_args()
        assert args[0] == "80"

    def test_trim_effect(self):
        """Trim effect creates correct args."""
        t = fx.Trim(start=1.5, end=10.0)
        assert t.name == "trim"
        assert t.to_args() == ["1.5", "=10.0"]

        t = fx.Trim(start=0, duration=5.0)
        assert t.to_args() == ["0", "5.0"]

    def test_trim_validation(self):
        """Trim effect validates parameters."""
        with pytest.raises(ValueError):
            fx.Trim(start=0, end=10, duration=5)

    def test_rate_effect(self):
        """Rate effect creates correct args."""
        r = fx.Rate(sample_rate=48000)
        assert r.name == "rate"
        assert "-h" in r.to_args()
        assert "48000" in r.to_args()

    def test_channels_effect(self):
        """Channels effect creates correct args."""
        c = fx.Channels(channels=1)
        assert c.name == "channels"
        assert c.to_args() == ["1"]

    def test_highpass_effect(self):
        """HighPass effect creates correct args."""
        h = fx.HighPass(frequency=80)
        assert h.name == "highpass"
        assert h.to_args() == ["-2", "80"]

    def test_highpass_validation(self):
        """HighPass effect validates parameters."""
        with pytest.raises(ValueError):
            fx.HighPass(frequency=100, poles=3)

    def test_fade_effect(self):
        """Fade effect creates correct args."""
        f = fx.Fade(fade_in=0.5, fade_out=1.0)
        assert f.name == "fade"
        args = f.to_args()
        assert "t" in args  # default type
        assert "0.5" in args

    def test_normalize_effect(self):
        """Normalize effect creates correct args."""
        n = fx.Normalize()
        assert n.name == "norm"
        assert n.to_args() == ["-1"]

        n = fx.Normalize(level=-3)
        assert n.to_args() == ["-3"]

    def test_effect_repr(self):
        """Effect classes have useful repr."""
        v = fx.Volume(db=3)
        assert "Volume" in repr(v)
        assert "db=3" in repr(v)


class TestCompositeEffect:
    """Tests for CompositeEffect."""

    def test_composite_effect_expand(self):
        """CompositeEffect can be defined and expanded."""
        class WarmReverb(fx.CompositeEffect):
            def __init__(self, decay=60):
                self.decay = decay

            @property
            def effects(self):
                return [
                    fx.HighPass(frequency=80),
                    fx.Reverb(reverberance=self.decay),
                    fx.Volume(db=-2),
                ]

        warm = WarmReverb(decay=70)
        assert len(warm.effects) == 3
        assert isinstance(warm.effects[0], fx.HighPass)
        assert isinstance(warm.effects[1], fx.Reverb)
        assert isinstance(warm.effects[2], fx.Volume)

    def test_composite_effect_to_args_raises(self):
        """CompositeEffect.to_args() raises TypeError."""
        class TestComposite(fx.CompositeEffect):
            @property
            def effects(self):
                return [fx.Volume(db=0)]

        c = TestComposite()
        with pytest.raises(TypeError):
            c.to_args()


class TestConcat:
    """Tests for cysox.concat()."""

    def test_concat_two_files(self, test_wav_str, output_path):
        """concat() joins two files."""
        cysox.concat([test_wav_str, test_wav_str], output_path)

        # Output should exist and have roughly 2x duration
        input_info = cysox.info(test_wav_str)
        output_info = cysox.info(str(output_path))

        assert output_info["sample_rate"] == input_info["sample_rate"]
        assert output_info["channels"] == input_info["channels"]
        # Allow 10% tolerance for duration
        expected_duration = input_info["duration"] * 2
        assert 0.9 <= output_info["duration"] / expected_duration <= 1.1

    def test_concat_three_files(self, test_wav_str, output_path):
        """concat() joins three files."""
        cysox.concat([test_wav_str, test_wav_str, test_wav_str], output_path)

        input_info = cysox.info(test_wav_str)
        output_info = cysox.info(str(output_path))

        expected_duration = input_info["duration"] * 3
        assert 0.9 <= output_info["duration"] / expected_duration <= 1.1

    def test_concat_single_file_raises(self, test_wav_str, output_path):
        """concat() raises ValueError for single file."""
        with pytest.raises(ValueError, match="at least 2"):
            cysox.concat([test_wav_str], output_path)

    def test_concat_empty_list_raises(self, output_path):
        """concat() raises ValueError for empty list."""
        with pytest.raises(ValueError, match="at least 2"):
            cysox.concat([], output_path)

    def test_concat_nonexistent_file_raises(self, test_wav_str, output_path):
        """concat() raises exception for nonexistent file."""
        with pytest.raises(Exception):
            cysox.concat([test_wav_str, "/nonexistent/file.wav"], output_path)


class TestPlay:
    """Tests for cysox.play()."""

    @pytest.mark.skip(reason="Requires audio hardware - run manually with: pytest -k test_play --run-audio")
    def test_play_basic(self, test_wav_str):
        """play() plays audio to default device."""
        cysox.play(test_wav_str, effects=[fx.Trim(start=0, duration=0.5)])

    @pytest.mark.skip(reason="Requires audio hardware - run manually with: pytest -k test_play --run-audio")
    def test_play_with_effects(self, test_wav_str):
        """play() applies effects during playback."""
        cysox.play(test_wav_str, effects=[
            fx.Trim(start=0, duration=0.5),
            fx.Volume(db=-6),
        ])

    def test_play_nonexistent_file(self):
        """play() raises exception for nonexistent file."""
        with pytest.raises(Exception):
            cysox.play("/nonexistent/file.wav")


class TestAutoInit:
    """Tests for auto-initialization behavior."""

    def test_no_manual_init_required(self, test_wav_str):
        """High-level API works without manual init()."""
        # This test verifies that we can use the API without calling sox.init()
        info = cysox.info(test_wav_str)
        assert info["sample_rate"] > 0

    def test_low_level_api_accessible(self):
        """Low-level API is accessible via cysox.sox."""
        from cysox import sox
        assert hasattr(sox, "Format")
        assert hasattr(sox, "init")
        assert hasattr(sox, "quit")
